# flake8: noqa
from ._logging import ConsoleErrHandler as ConsoleErrHandler
from ._logging import ConsoleHandler as ConsoleHandler
from ._logging import LogLevel as LogLevel
from ._logging import LogPublisher as LogPublisher
from ._logging import Logger as Logger
from ._logging import global_log_publisher as global_log_publisher



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk303_hgr_filler_plotter

Created: 2017/02/17

Description:
    Macro that illustrates how to loop over multiple (possibly large!)
    datasets in chunks, in each loop fill a histogrammar histograms,
    and plot the final histograms.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import analysis, core_ops, process_manager, resources, visualization, ConfigObject, Chain
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk303_hgr_filler_plotter.py.')

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk303_hgr_filler_plotter'
settings['version'] = 0

#########################################################################################

msg = r"""

The plots and latex files produced by link hist_summary can be found in dir:
{path}
"""
logger.info(msg, path=settings['resultsDir'] + '/' + settings['analysisName'] + '/data/v0/report/')

# --- Analysis configuration flags.
#     E.g. use these flags turn on or off certain chains with links.
#     by default all set to false, unless already configured in
#     configobject or vars()

settings['do_loop'] = True
settings['do_plotting'] = True

chunk_size = 400

#########################################################################################
# --- create dummy example dataset, which is read in below

input_files = [resources.fixture('mock_accounts.csv.gz')]


def to_date(x):
    """Convert to timestamp."""
    import pandas as pd
    try:
        ts = pd.Timestamp(x.split()[0])
        return ts
    except BaseException:
        pass
    return x


#########################################################################################
# --- now set up the chains and links, based on configuration flags

# --- example 2: readdata loops over the input files, with file chunking.

if settings['do_loop']:
    ch = Chain('Data')

    # --- a loop is set up in the chain MyChain.
    #     we iterate over (chunks of) the next file in the list until the iterator is done.
    #     then move on to the next chain (Overview)

    # --- readdata keeps on opening the next 400 lines of the open or next file in the file list.
    #     all kwargs are passed on to pandas file reader.
    read_data = analysis.ReadToDf(name='dflooper', key='rc', reader='csv')
    read_data.chunksize = chunk_size
    read_data.path = input_files
    ch.add(read_data)

    # add conversion functions to "Data" chain
    # here, convert column 'registered', an integer, to an actual timestamp.
    conv_funcs = [{'func': to_date, 'colin': 'registered', 'colout': 'date'}]
    transform = analysis.ApplyFuncToDf(name='Transform', read_key=read_data.key,
                                       apply_funcs=conv_funcs)
    ch.add(transform)

    # --- As an example, will fill histogram iteratively over the file loop
    hf = analysis.HistogrammarFiller()
    hf.read_key = 'rc'
    hf.store_key = 'hist'
    hf.logger.log_level = LogLevel.DEBUG
    # colums that are picked up to do value_counting on in the input dataset
    # note: can also be 2-dim: ['isActive','age']
    # in this example, the rest are one-dimensional histograms
    hf.columns = ['date', 'isActive', 'age', 'eyeColor', 'gender', 'company',
                  'latitude', 'longitude', ['isActive', 'age'], ['latitude', 'longitude']]
    # binning is apply to all input columns that are numeric or timestamps.
    # default binning is: bin_width = 1, bin_offset = 0
    # for timestamps, default binning is: { 'bin_width': np.timedelta64(30,'D'),
    #                                       'bin_offset': np.datetime64('2010-01-04') } }
    hf.bin_specs = {'longitude': {'bin_width': 5, 'bin_offset': 0},
                    'latitude': {'bin_width': 5, 'bin_offset': 0}}
    ch.add(hf)

    # --- this serves as the continue statement of the loop. go back to start of the chain.
    repeater = core_ops.RepeatChain()
    # repeat until readdata says halt.
    repeater.listen_to = 'chainRepeatRequestBy_' + read_data.name
    ch.add(repeater)

    link = core_ops.DsObjectDeleter()
    link.keep_only = ['hist', 'n_sum_rc', 'rc']
    ch.add(link)

# --- print contents of the datastore
if settings['do_plotting']:
    ch = Chain('Overview')
    pds = core_ops.PrintDs(name='End')
    pds.keys = ['n_sum_rc']
    ch.add(pds)

    # --- make a nice summary report of the created histograms
    hist_summary = visualization.DfSummary(name='HistogramSummary',
                                           read_key=hf.store_key)
    ch.add(hist_summary)

#########################################################################################

logger.debug('Done parsing configuration file esk303_hgr_filler_plotter.py.')



"""Project: Eskapade - A python-based package for data analysis.

Class: WriteFromDf

Created: 2016/11/08

Description:
    Algorithm to write a DataFrame from the DataStore to disk

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy
import os

import pandas as pd

from eskapade import process_manager, DataStore, Link, StatusCode
from eskapade.core import persistence
from eskapade.logger import Logger

pd_writers = {'csv': pd.DataFrame.to_csv,
              'xls': pd.DataFrame.to_excel,
              'xlsx': pd.DataFrame.to_excel,
              'json': pd.DataFrame.to_json,
              'h5': pd.DataFrame.to_hdf,
              'sql': pd.DataFrame.to_sql,
              'htm': pd.DataFrame.to_html,
              'html': pd.DataFrame.to_html,
              'dta': pd.DataFrame.to_stata,
              'pkl': pd.DataFrame.to_pickle,
              'pickle': pd.DataFrame.to_pickle}
logger = Logger()


class WriteFromDf(Link):
    """Write a DataFrame from the DataStore to disk."""

    def __init__(self, **kwargs):
        """Store the configuration of the link.

        :param str name: Name given to the link
        :param str key: the DataStore key
        :param str path: path where to save the DataFrame
        :param writer: file extension that can be written by a pandas writer function from pd.DataFrame.
            For example: 'csv'
        :param dict dictionary: keys (as in the arg above) and paths (as in the arg above)
            it will write out all the keys to the associated paths.
        :param bool add_counter_to_name: if true, add an index to the output file name.
            Useful when running in loops. Default is false.
        :param kwargs: all other key word arguments are passed on to the pandas writers.
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'WriteFromDf'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs, path='', key='', writer=None, dictionary={}, add_counter_to_name=False)

        # pass on remaining kwargs to pandas writer
        self.kwargs = copy.deepcopy(kwargs)

        # execute counter
        self._counter = 0
        return

    def initialize(self):
        """Initialize the link."""
        # perform basic checks of configured attributes
        # a key and path OR dictionary need to have been set.
        if self.path and self.key:
            self.dictionary = {self.key: self.path}
        elif not self.dictionary:
            raise Exception('Path and key OR dictionary not properly set.')

        # correct the output paths, if need be
        paths = list(self.dictionary.values())
        assert '' not in paths, 'One or more of the paths in dict is empty.'
        assert all([isinstance(p, str) for p in paths]), 'One or more of the paths in dict is not string.'
        # update paths if needed
        for k, p in self.dictionary.items():
            if not p.__contains__('/'):
                self.dictionary[k] = persistence.io_path('results_data', p)
                self.logger.debug('Output filename for key <{key}> has been reset to {new_key}.',
                                  key=k, new_key=self.dictionary[k])
        self.logger.info('kwargs passed on to pandas writer are: {kwargs}.', kwargs=self.kwargs)

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Pick up the dataframe and write to disk.
        """
        ds = process_manager.service(DataStore)

        # check that all dataframes are present
        assert all(k in ds for k in self.dictionary), 'key(s) not in DataStore.'

        # check that all ds items are dataframes
        assert all(isinstance(ds[k], pd.DataFrame) for k in self.dictionary), \
            'key(s) is not a pandas DataFrame.'

        # collect writer and store the dataframes
        for k, path in self.dictionary.items():
            df = ds[k]
            if self.add_counter_to_name:
                ps = os.path.splitext(path)
                path = ps[0] + '_' + str(self._counter) + ps[1]
            writer = pandas_writer(path, self.writer)
            folder = os.path.dirname(path)
            self.logger.debug('Checking for directory <{dir}>.', dir=folder)
            if not os.path.exists(folder):
                self.logger.fatal('Path given is invalid.')
            self.logger.debug('Writing file "{path}".', path=path)
            writer(df, path, **self.kwargs)

        self._counter += 1
        return StatusCode.Success


def pandas_writer(path, writer):
    """Pick the correct pandas writer.

    Based on provided writer setting, or based on file extension.
    """
    if isinstance(writer, str):
        try:
            writer = pd_writers.get(writer)
        except Exception:
            ValueError('writer parameter should be an extension pandas can write to disk.')
    if not writer:
        writer = pd_writers.get(os.path.splitext(path)[1].strip('.'), None)
    if not writer:
        logger.fatal('No suitable writer found for file "{path}".', path=path)
        raise RuntimeError('Unable to find suitable Pandas writer.')
    logger.debug('Using Pandas writer "{writer!s}"', writer=writer)
    return writer



"""Project: Eskapade - A python-based package for data analysis.

Class : correlation_summary

Created: 2017/03/13

Description:
    Algorithm to do create correlation heatmaps.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import os

import numpy as np
import pandas as pd
import tabulate
from sklearn.feature_selection import mutual_info_regression

from eskapade import process_manager, resources, Link, DataStore, StatusCode
from eskapade import visualization
from eskapade.core import persistence

ALL_CORRS = ['pearson', 'kendall', 'spearman', 'correlation_ratio']
LINEAR_CORRS = ['pearson', 'kendall', 'spearman']


class CorrelationSummary(Link):
    """Create a heatmap of correlations between dataframe variables."""

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str read_key: key of input dataframe to read from data store
        :param str store_key: key of correlations dataframe in data store
        :param str results_path: path to save correlation summary pdf
        :param list methods: method(s) of computing correlations
        :param str pages_key: data store key of existing report pages
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'correlation_summary'))

        # process arguments
        self._process_kwargs(kwargs, read_key='', store_key='', results_path='', methods=ALL_CORRS, pages_key='')
        self.check_extra_kwargs(kwargs)

    def _process_results_path(self):
        """Process results_path argument."""
        if not self.results_path:
            self.results_path = persistence.io_path('results_data', 'report')
        persistence.create_dir(self.results_path)

    def initialize(self):
        """Initialize the link."""
        # check input arguments
        self.check_arg_types(read_key=str, store_key=str, results_path=str, methods=list, pages_key=str)
        self.check_arg_vals('read_key')

        # read report templates
        with open(resources.template('df_summary_report.tex')) as templ_file:
            self.report_template = templ_file.read()
        with open(resources.template('df_summary_report_page.tex')) as templ_file:
            self.page_template = templ_file.read()

        self._process_results_path()

        # check methods
        for method in self.methods:
            if method not in ALL_CORRS:
                logstring = '"{}" is not a valid correlation method, please use one of {}'
                logstring = logstring.format(method, ', '.join(['"' + m + '"' for m in ALL_CORRS]))
                raise AssertionError(logstring)

        # initialize attributes
        self.pages = []

        return StatusCode.Success

    def execute(self):
        """Execute the link."""
        ds = process_manager.service(DataStore)

        # fetch and check input data frame
        # drop all-nan columns right away
        df = ds.get(self.read_key, None).dropna(how='all', axis=1)
        if not isinstance(df, pd.DataFrame):
            self.logger.fatal('No Pandas data frame "{self.read_key}" found in data store for {self!s}.', self=self)
            raise RuntimeError('No input data found for {!s}.', self)
        n_df = len(df.index)
        assert n_df, 'Pandas data frame "{}" frame has zero length.'.format(self.read_key)

        # create report pages
        if self.pages_key:
            self.pages = ds.get(self.pages_key, [])
            assert isinstance(self.pages, list), 'Pages key {} does not refer to a list.'.format(self.pages_key)

        # below, create report pages
        # for each correlation create resulting heatmap
        cors_list = []

        for method in self.methods:
            # compute correlations between all numerical variables
            self.logger.debug('Computing "{method}" correlations of dataframe "{key}".',
                              method=method, key=self.read_key)

            # mutual info, from sklearn
            if method == 'mutual_information':
                # numerical columns only
                cols = df.select_dtypes(include=[np.number]).columns

                # initialize correlation matrix
                n = len(cols)
                cors = np.zeros((n, n))
                for i, c in enumerate(cols):
                    # compare each column to all of the columns
                    cors[i, :] = mutual_info_regression(df[cols], df[c])

                cors = pd.DataFrame(cors, columns=cols, index=cols)

            elif method == 'correlation_ratio':
                # numerical columns only
                cols = df.select_dtypes(include=[np.number]).columns

                # choose bins for each column
                bins = {c: len(np.histogram(df[c])[1]) for c in cols}

                # sort rows into bins
                for c in cols:
                    df[str(c) + '_bin'] = pd.cut(df[c], bins[c])

                # initialize correlation matrix
                n = len(cols)
                cors = np.zeros((n, n))

                for i, x in enumerate(cols):
                    # definition from Wikipedia "correlation ratio"
                    xbin = str(x) + '_bin'
                    y_given_x = (df.groupby(xbin))[cols]
                    weighted_var_y_bar = (y_given_x.count() * (y_given_x.mean() - df.mean()) ** 2).sum()
                    weighted_var_y = df[cols].count() * df[cols].var()
                    cors[i, :] = weighted_var_y_bar / weighted_var_y

                cors = pd.DataFrame(cors, columns=cols, index=cols)

            else:
                cors = df.corr(method=method)
                cols = list(cors.columns)

            # replace column names with indices, as with numpy matrix, for plotting function below
            n = len(cols)
            cors.columns = range(n)

            # keep for potential later usage
            cors_list.append(cors)

            # plot settings
            title = '{0:s} correlation matrix'.format(method.capitalize())
            vmin = -1 if method in LINEAR_CORRS else 0
            vmax = 1
            color_map = 'RdYlGn' if method in LINEAR_CORRS else 'YlGn'
            fname = '_'.join(['correlations', self.read_key.replace(' ', ''), method]) + '.pdf'
            fpath = os.path.join(self.results_path, fname)

            # create nice looking plot
            self.logger.debug('Saving correlation heatmap as "{path}".', path=fpath)
            visualization.vis_utils.plot_correlation_matrix(cors, cols, cols, fpath, title, vmin, vmax, color_map)

            # statistics table for report page
            n_unique = (n * n - n) / 2 if method is not 'correlation_ratio' else n * n
            stats = [('entries', n_df), ('bins', n * n), ('unique', n_unique),
                     ('> 0', (cors.values.ravel() > 0).sum()),
                     ('< 0', (cors.values.ravel() < 0).sum()),
                     ('avg', np.average(cors.values.ravel())),
                     ('max', max(cors.values.ravel())),
                     ('min', min(cors.values.ravel()))] if n > 0 else []
            stats_table = tabulate.tabulate(stats, tablefmt='latex')

            # add plot and table as page to report
            self.pages.append(self.page_template.replace('VAR_LABEL', title)
                              .replace('VAR_STATS_TABLE', stats_table)
                              .replace('VAR_HISTOGRAM_PATH', fpath))

        # save correlations to datastore if requested
        if self.store_key:
            ds[self.store_key] = cors_list
        if self.pages_key:
            ds[self.pages_key] = self.pages

        return StatusCode.Success

    def finalize(self):
        """Finalize the link."""
        # write report file
        with open('{}/report.tex'.format(self.results_path), 'w') as report_file:
            report_file.write(
                self.report_template.replace(
                    'INPUT_PAGES', ''.join(
                        self.pages)))

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk305_correlation_summary

Created: 2017/04/04

Description:
    Macro to demonstrate generating correlation heatmaps


Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, resources, Chain
from eskapade import analysis, visualization
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk305_correlation_summary.')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk305_correlation_summary'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

settings['input_path'] = resources.fixture('correlated_data.sv.gz')
settings['reader'] = 'csv'
settings['separator'] = ' '
settings['correlations'] = ['pearson', 'kendall', 'spearman', 'correlation_ratio']

#########################################################################################
# --- now set up the chains and links based on configuration flags

# create chains
data = Chain('Data')

# load data
reader = analysis.ReadToDf(name='reader',
                           path=settings['input_path'],
                           sep=settings['separator'],
                           key='input_data',
                           reader=settings['reader'])

data.add(reader)

# make visualizations of correlations
summary = Chain('Summary')

corr_link = visualization.CorrelationSummary(name='correlation_summary',
                                             read_key='input_data',
                                             store_key='correlations',
                                             methods=settings['correlations'])

summary.add(corr_link)

#########################################################################################

logger.debug('Done parsing configuration file esk305_correlation_summary.')



"""Project: Eskapade - A python-based package for data analysis.

Class: ToDsDict

Created: 2016/11/08

Description:
    Algorithm to store one object in the DataStore dict during run time.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class ToDsDict(Link):
    """Stores one object in the DataStore dict during run time."""

    def __init__(self, **kwargs):
        """Link to store one external object in the DataStore dict during run time.

        :param str name: name of link
        :param str store_key: key of object to store in data store
        :param obj: object to store
        :param bool force: overwrite if already present in datastore. default is false. (optional)
        :param bool at_initialize: store at initialize of link. Default is false.
        :param bool at_execute: store at execute of link. Default is true.
        :param bool copydict: if true and obj is a dict, copy all key value pairs into datastore. Default is false.
        """
        Link.__init__(self, kwargs.pop('name', 'ToDsDict'))

        # process keyword arguments
        self._process_kwargs(kwargs,
                             store_key=None,
                             obj=None,
                             at_initialize=False,
                             at_execute=True,
                             force=False,
                             copydict=False)
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link."""
        # perform basic checks.
        if self.obj is None:
            raise RuntimeError('object "{}" to store is of type None'.format(self.store_key))
        # storage key needs to be set in nearly all cases
        if not (self.copydict and isinstance(self.obj, dict)):
            if not (isinstance(self.store_key, str) and self.store_key):
                raise RuntimeError('object storage key has not been set')

        ds = process_manager.service(DataStore)
        return StatusCode.Success if not self.at_initialize else self.do_storage(ds)

    def execute(self):
        """Execute the link."""
        ds = process_manager.service(DataStore)

        return StatusCode.Success if not self.at_execute else self.do_storage(ds)

    def do_storage(self, ds):
        """Perform storage in datastore.

        Function makes a distinction been dicts and any other object.
        """
        # if dict and copydict==true, store all individual items
        if self.copydict and isinstance(self.obj, dict):
            stats = [(self.store(ds, v, k, force=self.force)).value for k, v in self.obj.items()]
            return StatusCode(max(stats))

        # default: store obj under store_key
        return self.store(ds, self.obj, force=self.force)



"""Project: Eskapade - A python-based package for data analysis.

Created: 2016/11/08

Description:
    Utility class and functions to get correct io path,
    used for persistence of results

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import glob
import os
import re
from collections import defaultdict

from eskapade.core.process_manager import process_manager
from eskapade.core.process_services import ConfigObject
from eskapade.logger import Logger

# IO locations
IO_LOCS = dict(config='config_dir',
               results='results_dir',
               data='data_dir',
               macros='macros_dir',
               input_data='data_dir',
               records='data_dir',
               ana_results='results_dir',
               ana_plots='results_dir',
               proc_service_data='results_dir',
               results_data='results_dir',
               results_ml_data='results_dir',
               results_config='results_dir',
               tmva='results_dir',
               plots='results_dir',
               templates='templates_dir')

IO_SUB_DIRS = defaultdict(lambda: '',
                          ana_results='{ana_name:s}',
                          ana_plots='{ana_name:s}/plots',
                          proc_service_data='{ana_name:s}/proc_service_data/v{ana_version:s}',
                          results_data='{ana_name:s}/data/v{ana_version:s}',
                          results_ml_data='{ana_name:s}/data/v{ana_version:s}',
                          results_config='{ana_name:s}/config/v{ana_version:s}',
                          tmva='{ana_name:s}/tmva_output/v{ana_version:s}',
                          plots='{ana_name:s}/plots/v{ana_version:s}')

# get logging instance
logger = Logger()


def repl_whites(name):
    """Replace whitespace in names."""
    return '_'.join(name.split())


def create_dir(dir_path):
    """Create directory.

    :param str dir_path: directory path
    """
    if os.path.exists(dir_path):
        if os.path.isdir(dir_path):
            return
        logger.fatal('Directory path "{path}" exists, but is not a directory.', path=dir_path)
        raise AssertionError('Unable to create IO directory.')
    logger.debug('Creating directory "{path}".', path=dir_path)
    os.makedirs(dir_path)


def io_dir(io_type, io_conf=None):
    """Construct directory path.

    :param str io_type: type of result to store, e.g. data, macro, results.
    :param io_conf: IO configuration object
    :return: directory path
    :rtype: str
    """
    if not io_conf:
        io_conf = process_manager.service(ConfigObject).io_conf()
    # check inputs
    if io_type not in IO_LOCS:
        logger.fatal('Unknown IO type: "{type!s}".', type=io_type)
        raise RuntimeError('IO directory found for specified IO type.')
    if IO_LOCS[io_type] not in io_conf:
        logger.fatal('Directory for io_type ({type}->{path}) not in io_conf.', type=io_type, path=IO_LOCS[io_type])
        raise RuntimeError('io_dir: directory for specified IO type not found in specified IO configuration.')

    # construct directory path
    base_dir = io_conf[IO_LOCS[io_type]]
    sub_dir = IO_SUB_DIRS[io_type].format(ana_name=repl_whites(io_conf['analysis_name']),
                                          ana_version=repl_whites(str(io_conf['analysis_version'])))
    dir_path = base_dir + ('/' if base_dir[-1] != '/' else '') + sub_dir

    # create and return directory path
    create_dir(dir_path)
    return dir_path


def io_path(io_type, sub_path, io_conf=None):
    """Construct directory path with sub path.

    :param str io_type: type of result to store, e.g. data, macro, results.
    :param str sub_path: sub path to be included in io path
    :param io_conf: IO configuration object
    :return: full path to directory
    :rtype: str
    """
    if not io_conf:
        io_conf = process_manager.service(ConfigObject).io_conf()
    # check inputs
    if not isinstance(sub_path, str):
        logger.fatal('Specified sub path/file name must be a string, but has type "{type!s}"',
                     type=type(sub_path).__name__)
        raise TypeError('The sub path/file name in the io_path function must be a string')
    sub_path = repl_whites(sub_path).strip('/')

    # construct path
    full_path = io_dir(io_type, io_conf) + '/' + sub_path
    if os.path.dirname(full_path):
        create_dir(os.path.dirname(full_path))

    return full_path


def record_file_number(file_name_base, file_name_ext, io_conf=None):
    """Get next prediction-record file number.

    :param str file_name_base: base file name
    :param str file_name_ext: file name extension
    :param io_conf: I/O configuration object
    :return: next prediction-record file number
    :rtype: int
    """
    if not io_conf:
        io_conf = process_manager.service(ConfigObject).io_conf()
    file_name_base = repl_whites(file_name_base)
    file_name_ext = repl_whites(file_name_ext)
    records_dir = io_dir('records', io_conf)
    max_num = -1
    regex = re.compile('.*/{}_(\d*?).{}'.format(file_name_base, file_name_ext))
    for file_path in glob.glob('{}/{}_*.{}'.format(records_dir, file_name_base, file_name_ext)):
        digit = regex.search(file_path)
        if digit:
            max_num = max(max_num, int(digit.group(1)))

    return max_num + 1



"""Project: Eskapade - A python-based package for data analysis.

Class: DataFrameColumnRenamer

Created: 2016/11/08

Description:
    Algorithm to concatenate multiple pandas datadrames

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy

import pandas as pd

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class DfConcatenator(Link):
    """Concatenates multiple pandas datadrames."""

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str store_key: key of data to store in data store
        :param list read_keys: keys of pandas dataframes in the data store
        :param bool ignore_missing_input: Skip missing input datasets.
            If all missing, store empty dataset. Default is false.
        :param kwargs: all other key word arguments are passed on to pandas concat function.
        """
        Link.__init__(self, kwargs.pop('name', 'DfConcatenator'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs, read_keys=[])
        self._process_kwargs(kwargs, store_key=None)
        self._process_kwargs(kwargs, ignore_missing_input=False)

        # pass on remaining kwargs to pandas reader
        self.kwargs = copy.deepcopy(kwargs)

    def initialize(self):
        """Initialize the link."""
        assert self.read_keys, 'read_keys have not been set.'
        assert isinstance(self.store_key, str) and self.store_key, 'storage key not set.'

        self.logger.info('kwargs passed on to pandas concat function are: {kwargs}', kwargs=self.kwargs)

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Perform concatenation of multiple pandas datadrames.
        """
        ds = process_manager.service(DataStore)

        # check if all input dataframes exist. if so configured, skip missing inputs, else raise e.
        data = []
        for c in self.read_keys:
            if c not in ds:
                if self.ignore_missing_input:
                    self.logger.warning("<{key}> is not a key in the datastore. Configured to skip it.", key=c)
                    continue
                raise Exception("<{}> is not a key in the datastore.".format(c))
            data.append(ds[c])

        # concatenate the dataframes
        if len(data):
            df = pd.concat(data, **self.kwargs).reset_index(drop=True)
        elif self.ignore_missing_input:
            self.logger.warning("Nothing to concatenate. Configured to return empty dataframe.")
            df = pd.DataFrame()
        else:
            raise Exception("Nothing to concatenate.")

        # store the result
        ds[self.store_key] = df
        ds['n_' + self.store_key] = len(df.index)

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk208_record_factorizer

Created: 2017/03/13

Description:
    This macro performs the factorization of an input column of an input
    dataframe.  E.g. a columnn x with values 'apple', 'tree', 'pear',
    'apple', 'pear' is tranformed into columns x with values 0, 1, 2, 0, 2.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, resources, Chain
from eskapade import core_ops, analysis
from eskapade import process_manager
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk208_record_factorizer')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk208_record_factorizer'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# --- Set path of data
data_path = resources.fixture('dummy.csv')

#########################################################################################
# --- now set up the chains and links based on configuration flags

factorize = Chain('Factorize')

# --- read dummy dataset
read_data = analysis.ReadToDf(key='test1', sep='|', reader='csv', path=data_path)
factorize.add(read_data)

# --- print contents of the datastore
pds = core_ops.PrintDs(name='printer1')
pds.keys = ['test1']
factorize.add(pds)

# --- add the record factorizer
#     Here the columns dummy and loc of the input dataset are factorized
#     e.g. x = ['apple', 'tree', 'pear', 'apple', 'pear'] becomes the column:
#     x = [0, 1, 2, 0, 2]
#     By default, the mapping is stored in a dict under key: 'map_'+store_key+'_to_original'
fact = analysis.RecordFactorizer(name='rf1')
fact.columns = ['dummy', 'loc']
fact.read_key = 'test1'
fact.store_key = 'test1_fact'
fact.sk_map_to_original = 'to_original'
fact.logger.log_level = LogLevel.DEBUG
factorize.add(fact)

# --- print contents of the datastore
pds = core_ops.PrintDs(name='printer2')
pds.keys = ['to_original', 'test1', 'test1_fact']
factorize.add(pds)

refactorize = Chain('ReFactorize')

# --- add second record factorizer, which now maps the columns
#     dummy and loc back to their original format
factorizer = analysis.RecordFactorizer(name='rf2')
factorizer.read_key = fact.store_key
factorizer.store_key = 'test1_refact'
factorizer.map_to_original = fact.sk_map_to_original
factorizer.logger.log_level = LogLevel.DEBUG
refactorize.add(factorizer)

# --- print contents of the datastore
pds = core_ops.PrintDs(name='printer3')
pds.keys = ['test1', 'test1_fact', 'test1_refact']
refactorize.add(pds)

#########################################################################################

logger.debug('Done parsing configuration file esk208_record_factorizer')



"""Project: Eskapade - A python-based package for data analysis.

Class: RandomSampleSplitter

Created: 2016/11/08

Description:
    Algorithm to randomly assign records to a number of classes

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import pandas as pd
from numpy.random import RandomState

from eskapade import process_manager, ConfigObject, DataStore, Link, StatusCode


class RandomSampleSplitter(Link):
    """Link that randomly assigns records of an input dataframe to a number of classes.

    After assigning classes does one of the following:

    - splits the input dataframe into sub dataframes according classes and stores the sub dataframes into the datastore;
    - add a new column with assigned classes to the dataframe.

    Records are assigned randomly.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str read_key: key of data to read from datastore
        :param list store_key: keys of datasets to store in datastore.
            Number of sub samples equals length of store_key list (optional instead of 'column' and 'nclasses').
        :param str column: name of new column that specifies the randomly assigned class.
            Default is randomclass (optional instead of 'store_key').
        :param int nclasses: number of random classes. Needs to be set
            (optional instead of 'store_key').
        :param list fractions: list of fractions (0<fraction<1) of records assigned to the sub samples.
            Can be one less than n classes. Sum can be less than 1. Needs to be set.
        :param list nevents: list of number of random records assigned to the sub samples
            Can be one less than n classes (optional instead of 'fractions').
        """
        Link.__init__(self, kwargs.pop('name', 'RandomSampleSplitter'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             read_key=None,
                             store_key=None,
                             column='randomclass',
                             fractions=None,
                             nevents=False,
                             nclasses=None)

        # check residual kwargs. exit if any present.
        self.check_extra_kwargs(kwargs)

    def _convert_attr_to_list(self, var_name):
        """Convert a float attribute to a list.

        If the attribute is a single float, converts it to a list.
        If the attribute is neither list nor float, throws an exception.
        :param var_name: Attribute name.
        """
        var = getattr(self, var_name)
        if isinstance(var, float):
            setattr(self, var_name, [var])
        elif not isinstance(var, list):
            raise Exception('Given {var_name} of incorrect type.'.format(var_name=var_name))

    def initialize(self):
        """Check and initialize attributes of the link."""
        # check that these four attributes have been set correctly.
        assert isinstance(self.read_key, str) and self.read_key, 'read key not set.'
        # either store keys or nclasses and column need to be specified.
        if self.store_key is not None:
            if isinstance(self.store_key, str) and self.store_key:
                self.store_key = [self.store_key]
            elif not (isinstance(self.store_key, list) and len(self.store_key)):
                raise Exception('Store keys not filled with list of strings.')
            self.nclasses = len(self.store_key)
        else:
            assert isinstance(self.column, str) and self.column, 'Column not set.'
            assert isinstance(self.nclasses, int) and self.nclasses > 1, \
                'Number of random classes need to be int and greater than 1.'

        assert isinstance(self.nclasses, int) and self.nclasses >= 1, \
            'number of random classes need to be int and greater than or equal to 1'
        if self.fractions is None and self.nevents is None:
            raise Exception('No fractions or nevents are provided. Provide at least one.')

        # check that provided fractions are all okay.
        if self.fractions is not None:
            self._convert_attr_to_list('fractions')

            if len(self.fractions) == self.nclasses - 1:
                sumf = sum(f for f in self.fractions)
                self.fractions.append(1. - sumf)
            assert (len(self.fractions) == self.nclasses), \
                'Number of fractions <{:d}> doesnt equal number of classes <{:d}>.'.format(
                    len(self.fractions), self.nclasses)
            for i, f in enumerate(self.fractions):
                assert f >= 0, '<{:d}> assigned fraction <{:f}> needs to be greater than zero.'.format(i, f)
            # normalize fractions
            sumf = sum(f for f in self.fractions)
            for i, f in enumerate(self.fractions):
                if sumf > 1:
                    self.fractions[i] = f / sumf
                self.logger.info('Random class <{index:d}> assigned fraction is <{fraction:f}>.',
                                 index=i, fraction=self.fractions[i])

        # alternatively, check that provided number of events per random class are all okay.
        if self.nevents is not None:
            self._convert_attr_to_list('nevents')
            if not ((len(self.nevents) == self.nclasses) or (len(self.nevents) == self.nclasses - 1)):
                raise Exception('number of provided events <{:d}> does not equal number of classes <{:d}>.'
                                .format(len(self.nevents), self.nclasses))

        # there needs to be a random seed set in the configobject
        settings = process_manager.service(ConfigObject)
        assert 'seed' in settings, 'random seed not set in ConfigObject.'
        self._seed = settings['seed']

        return StatusCode.Success

    def execute(self):
        """Execute the link."""
        ds = process_manager.service(DataStore)

        # basic checks on contensts of the data frame
        assert self.read_key in ds, 'Key "{key}" not in DataStore.'.format(key=self.read_key)
        df = ds[self.read_key]
        if not isinstance(df, pd.DataFrame):
            raise Exception('Retrieved object not of type pandas DataFrame.')
        ndf = len(df.index)
        assert ndf > 0, 'dataframe {} is empty.'.format(self.read_key)
        if self.store_key is None:
            if self.column in df.columns:
                raise Exception('Column name <{}> already used: <{!s}>. Will not overwrite.'
                                .format(self.column, df.columns))
            df[self.column] = 0

        # fix final number of events assigned per random class
        # ... each class gets at least one event
        if self.nevents is not None:
            if len(self.nevents) == self.nclasses - 1:
                self.nevents.append(ndf - sum(n for n in self.nevents))
        else:
            self.nevents = [int(ndf * f) for f in self.fractions]
        for i in range(self.nclasses):
            nsum = sum(n for n in self.nevents[:i + 1])
            ndiff = 0 if nsum - ndf < 0 else nsum - ndf
            self.nevents[i] -= ndiff
            if self.nevents[i] < 0:
                self.nevents[i] = 0
            self.logger.info('Random class <{index:d}> assigned <{n:d}> events.', index=i, n=self.nevents[i])

        # random reshuffling of dataframe indices
        RNG = RandomState(self._seed)
        permute = RNG.permutation(df.index)

        # apply the random reshuffling, and assign records to the n datasets
        for i in range(self.nclasses):
            ib = sum(n for n in self.nevents[:i])
            ie = sum(n for n in self.nevents[:i + 1])
            if self.store_key is None:
                df.ix[permute[ib:ie], self.column] = i
            else:
                ds[self.store_key[i]] = df.ix[permute[ib:ie]]
                self.logger.info('Stored output collection <{key}> with <{n:d}> records in datastore.',
                                 key=self.store_key[i], n=len(ds[self.store_key[i]].index))

        # increase seed in case of next iteration
        self._seed += 1

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Class: HelloWorld

Created: 2017/01/31

Description:
    Algorithm to do print Hello {}!

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import Link
from eskapade import StatusCode


class HelloWorld(Link):
    """Defines the content of link HelloWorld."""

    def __init__(self, **kwargs):
        """Store the configuration of link HelloWorld.

        :param str name: name assigned to the link
        :param str hello: name to print in Hello World! Defaults to 'World'
        :param int repeat: repeat print statement N times. Default is 1
        """
        # initialize Link
        Link.__init__(self, kwargs.pop('name', 'HelloWorld'))

        # process keyword arguments
        self._process_kwargs(kwargs, hello='World', repeat=1)

        # check residual kwargs.
        # (turn line off if you wish to keep these to pass on.)
        self.check_extra_kwargs(kwargs)
        # self.kwargs = kwargs

    def execute(self):
        """Execute the link."""
        for _ in range(self.repeat):
            self.logger.info('Hello {hello}', hello=self.hello)

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk109_debugging_tips

Created: 2017/02/26

Description:
    Macro to illustrate basic debugging features of Eskapade.
    The macro shows how to start a python sessions while
    running through the chains, and also how to break out of a chain.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import DataStore
from eskapade import core_ops
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk109_debugging_tips')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk109_debugging_tips'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

msg = r"""

To end the eskapade_run session with a python console,
from the cmd line use the this flag: -i
"""
logger.info(msg)

# testing false used for running integration tests. do not remove.
settings['TESTING'] = settings.get('TESTING', False)

#########################################################################################
# --- now set up the chains and links based on configuration flags

ds = process_manager.service(DataStore)
ds['hello'] = 'world'
ds['d'] = {'a': 1, 'b': 2, 'c': 3}

#########################################################################################
# --- now set up the chains and links based on configuration flags

ch = Chain('Overview')

# 1. printdatastore prints an overview of the contents in the datastore
# at the state of executing the link.
# The overview consists of list of keys in the datastore and and the object types.
link = core_ops.PrintDs(name='printer1')
# keys are the items for which the contents of the actual item is printed.
link.keys = ['hello', 'd']
ch.add(link)

# 2. This link will start a python session.
# from this session, one can access the datastore and the configobject with:
# >>> ds
# or
# >>> settings
# Try to add something to the datastore in this session!
# >>> ds['foo'] = 'bar'
if not settings['TESTING']:
    link = core_ops.IPythonEmbed()
    ch.add(link)

# 3. let's see what has been added to the datastore ...
link = core_ops.PrintDs(name='printer2')
# keys are the items for which the contents of the actual item is printed.
link.keys = ['foo', 'hello', 'd']
ch.add(link)

# 4. print keys in store
def print_store(ds):
    print (ds.keys())
link = core_ops.DsApply()
link.apply = [print_store]
ch.add(link)

# 5. This link sends out a break signal!
# eskapade execution or any remaining links and chains is skipped.
link = core_ops.Break()
# keys are the items for which the contents of the actual item is printed.
ch.add(link)
link.keys = ['foo', 'hello', 'd']

# 6. this link should not be reached because of the Break!
ch = Chain('End')
link = core_ops.PrintDs(name='printer3')
ch.add(link)

# 7. eskapade_run with cmd line option -i to end the eskapade session with a python shell

#########################################################################################

logger.debug('Done parsing configuration file esk109_debugging_tips')



"""Project: Eskapade - A python-based package for data analysis.

Classes: ValueCounts, BinningUtil, Histogram

Created: 2017/03/14

Description:
    Generic 1D Histogram class.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

modification, are permitted according to the terms listed in the file
Redistribution and use in source and binary forms, with or without
LICENSE.
"""

from collections import Counter

import numpy as np
import pandas as pd
from sortedcontainers import SortedDict

from eskapade.core.mixin import ArgumentsMixin
from eskapade.logger import Logger


def _check_num_vals(cnts_iter):
    """Check all values in provided array are numeric.

    :raises TypeError: if any non-numeric type is found in input array
    """
    non_num = False
    for val in cnts_iter:
        if isinstance(val, str):
            non_num = True
            break
        try:
            float(val)
        except Exception:
            non_num = True
            break
    if non_num:
        raise TypeError('non-numeric type found ("{}")'.format(type(val).__name__))


class ValueCounts(object):
    """A dictionary of value counts.

    The dictionary of value counts comes out of pandas.series.value_counts()
    for one variable or pandas.Dataframe.groupby.size() performed over one
    or multiple variables.
    """

    def __init__(self, key, subkey=None, counts=None, sel=None):
        """Initialize link instance.

        :param list key: key is a tuple, list or string of (the) variable name(s), matching those and the structure of
               the keys in the value_counts dictionary.
        :param list subkey: subset of key. If provided, the value_counts dictionary will be projected from key onto the
               (subset of) subkey. E.g. use this to map a two dimensional value_counts dictionary onto one specified
               dimension. Default is None. Optional.
        :param dict counts: the value_counts dictionary.
        :param dict sel: Apply selections to value_counts dictionary. Default is {}. Optional.
        """
        if not counts:
            raise AttributeError('counts is not specified.')
        if sel is None:
            sel = {}
        key = self._transform_key(key)
        subkey = self._transform_key(subkey) if subkey is not None else key
        counts = dict((k if isinstance(k, tuple) else (k,), v) for k, v in counts.items())

        self._key = key
        self._skey = subkey if subkey is not None else key
        self._cnts = counts
        self._sel = dict((k, list(s) if hasattr(s, '__iter__') else [s]) for k, s in sel.items())
        self._ktos = tuple(key.index(k) for k in subkey)
        self._kind = dict((k, key.index(k)) for k in key)
        self._stok = tuple(subkey.index(k) if k in subkey else None for k in key)
        self._no_none_cnts = SortedDict()

    def __lt__(self, other):
        """Less than operator.

        :param object other: the other ValueCounts object
        :returns: true or false
        :rtype: bool
        """
        return len(self._key) < len(other._key)

    def __gt__(self, other):
        """Greater than operator.

        :param object other: the other ValueCounts object
        :returns: true or false
        :rtype: bool
        """
        return len(self._key) > len(other._key)

    def __eq__(self, other):
        """Equal to operator.

        :param object other: the other ValueCounts object
        :returns: true or false
        :rtype: bool
        """
        return len(self._key) == len(other._key) and tuple(self._skey) == tuple(other._skey)

    def __le__(self, other):
        """Less than or equal to operator.

        :param object other: the other ValueCounts object
        :returns: true or false
        :rtype: bool
        """
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other):
        """Greater than or equal to operator.

        :param object other: the other ValueCounts object
        :returns: true or false
        :rtype: bool
        """
        return self.__lt__(other) or self.__eq__(other)

    def _transform_key(self, key):
        """Transform input key to desired tuple format.

        Input key, a tuple, list, or string, gets transformed into the
        desired key format.  Desired key format is a tuple like this:

        * ('foo',) : for one variable (note the comma),
        * ('foo', 'bar') : for multiple variables.

        This format follows the same structure of the keys used in the
        internal value_counts dictionary.

        :param tuple key: input key, is a tuple, list, or string
        :returns: the tuplelized key
        :rtype: tuple
        """
        assert key, 'Input key contains no variable name(s). Expect str or tuple or list of strings.'

        has_itr = isinstance(key, (list, tuple))
        if has_itr:
            if len(key) == 1:
                key = (key[0],)
            else:
                key = tuple(key)
        elif isinstance(key, str):
            key = (key,)
        else:
            # don't recognize the key! pass.
            pass
        return key

    @property
    def counts(self):
        """Process and return value-counts dictionary.

        :returns: after processing, returns the value_counts dictionary
        :rtype: dict
        """
        self.process_counts()
        return self._cnts

    @property
    def nononecounts(self):
        """Return value-counts dictionary without None keys.

        :returns: after processing, returns the value_counts dictionary without None keys
        :rtype: dict
        """
        self.process_counts()
        if len(self._no_none_cnts) == 0:
            self._no_none_cnts = SortedDict([(var, cnt) for var, cnt in self._cnts.items() if None not in var])
        # return dict([(var, cnt) for var, cnt in self._cnts.items() if None
        # not in var])
        return self._no_none_cnts

    @property
    def key(self):
        """Process and return current value-counts key.

        :returns: the key
        :rtype: tuple
        """
        self.process_counts()
        return self._key

    @property
    def skey(self):  # noqa
        """Current value-counts subkey.

        :returns: the subkey
        :rtype: tuple
        """
        return self._skey

    @property
    def num_bins(self):  # noqa
        """Number of value-counts bins.

        :returns: number of bins
        :rtype: int
        """
        return len(self.counts)

    @property
    def num_nonone_bins(self):  # noqa
        """Number of not-none value-counts bins.

        :returns: number of not-none bins
        :rtype: int
        """
        return len(self.nononecounts)

    @property
    def sum_counts(self):  # noqa
        """Sum of counts of all value-counts bins.

        :returns: the sum of counts of all bins
        :rtype: float
        """
        return sum(self._cnts.values())

    @property
    def sum_nonone_counts(self):  # noqa
        """Sum of not-none counts of all value-counts bins.

        :returns: the sum of not-none counts of all bins
        :rtype: float
        """
        return sum(self.nononecounts.values())

    def create_sub_counts(self, subkey, sel=None):
        """Project existing value counts onto a subset of keys.

        E.g. map variables x,y onto single dimension x, so for each bin in x integrate over y.

        :param tuple subkey: input sub-key, is a tuple, list, or string.
                             This is the new key of variables for the returned ValueCounts object.
        :param dict sel: dictionary with selection. Optional.
        :returns: value_counts object where subkey has become the new key.
        :rtype: ValueCounts
        """
        subkey = self._transform_key(subkey)
        return ValueCounts(self.key, subkey, self.counts, sel)

    def count(self, value_bin):
        """Get bin count for specific bin-key value bin.

        :param tuple value_bin: a specific key, and can be a list or tuple.
        :returns: specific bin counter value
        :rtype: int
        """
        self.process_counts()
        return self._cnts.get(tuple(value_bin[k] for k in self._stok), 0)

    def get_values(self, val_keys=()):
        """Get all key-values of a subset of keys.

        E.g. give all x values in of the keys, when the value_counts object has keys (x, y).

        :param tuple val_keys: a specific sub-key to get key values for.
        :returns: all key-values of a subset of keys.
        :rtype: tuple
        """
        self.process_counts()
        if not val_keys:
            val_keys = self._skey
        return sorted(set(tuple(key[self._kind[k]] for k in val_keys) for key in self._cnts.keys()))

    def remove_keys_of_inconsistent_type(self, prefered_key_type=None):
        """Remove keys with inconsistent data type(s).

        :param tuple prefered_key_type: the prefered key type to keep. Can be a
                                        tuple, list, or single type.  E.g. str
                                        or (int, str, float).  If None provided,
                                        the most common key type found is kept.
        """
        self.process_counts()

        # NB: np.dtype(type(k).type : gives back a consistent numpy type for
        # all strings, floats, ints, etc.

        # convert prefered_key_type to right format
        if prefered_key_type is not None:
            # has_itr = hasattr(prefered_key_type, '__iter__')
            has_itr = isinstance(prefered_key_type, (list, tuple))

            if has_itr:
                if len(prefered_key_type) == 1:
                    prefered_key_type = (prefered_key_type[0],)
                else:
                    prefered_key_type = tuple(prefered_key_type)
            else:
                prefered_key_type = (prefered_key_type,)
            # turn into consistent types, used for comparison below
            prefered_key_type = tuple(np.dtype(k).type for k in prefered_key_type)

        # sort all keys by their key type, and count how often these types
        # occur
        cnts_types = Counter()
        cnts_keys = dict((tuple(np.dtype(type(k)).type for k in key), [])
                         for key in self._cnts)
        for key in self._cnts:
            ktype = tuple(np.dtype(type(k)).type for k in key)
            cnts_types[ktype] += self._cnts[key]
            cnts_keys[ktype].append(key)

        # pick the prefered key type to keep
        if prefered_key_type is None:
            # select most common key type
            prefered_key_type = cnts_types.most_common()[0][0]

        # remove all keys of different key type than preferred
        for ktype in cnts_types:
            if ktype == prefered_key_type:
                continue
            keys = cnts_keys[ktype]
            for k in keys:
                del self._cnts[k]

        # no_none_cnts gets refilled next time when called
        self._no_none_cnts.clear()

    def process_counts(self, accept_equiv=True):
        """Project value counts onto the existing subset of keys.

        E.g. map variables x,y onto single dimension x, so for each bin in x integrate over y.

        :param bool accept_equiv: accept equivalence of key and subkey if if
                                  subkey is in different order than key. Default
                                  is true.
        :returns: successful projection or not
        :rtype: bool
        """
        # only process if counts need processing
        if not self._sel and self._key == self._skey:
            return False
        if not self._sel and accept_equiv and all(k in self._skey for k in self._key):
            return False

        # create new counts dictionary with subcounts
        scnts = {}
        for vals, cnt in self._cnts.items():
            # apply selection
            if self._sel and any(vals[self._kind[k]] not in s for k, s in self._sel.items()):
                continue

            # add counts for subkey to sum
            vkey = tuple(vals[i] for i in self._ktos)
            if vkey not in scnts:
                scnts[vkey] = 0
            scnts[vkey] += cnt

        # set subcounts as new counts
        self._key = self._skey
        self._cnts = scnts
        self._sel = {}
        self._kind = dict((k, self._key.index(k)) for k in self._key)
        self._ktos = self._stok = tuple(range(len(self._skey)))
        # no_none_cnts refilled when called
        self._no_none_cnts.clear()
        return True


class BinningUtil(object):
    """Helper for interpreting bin specifications.

    BinningUtil is a helper class used for interpreting bin specification
    dictionaries.  It is a base class for the Histogram class.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        A bin_specs dictionary needs to be provided as input. bins_specs is
        a dict containing 'bin_width' and 'bin_offset' keys.  In case bins
        widths are not equal, bin_specs contains 'bin_edges' (array) instead
        of 'bin_width' and 'bin_offset'.  'bin_width' and 'bin_offset' can
        be numeric or numpy timestamps.

        Alternatively, bin_edges can be provided as input to bin_specs.

        Example bin_specs dictionaries are:

        >>> bin_specs = {'bin_width': 1, 'bin_offset': 0}
        >>> bin_spect = {'bin_edges': [0, 2, 3, 4, 5, 7, 8]}
        >>> bin_specs = {'bin_width': np.timedelta64(30, 'D'),
                         'bin_offset': np.datetime64('2010-01-04')}

        :param dict bin_specs: dictionary contains 'bin_width' and
                               'bin_offset' numbers or 'bin_edges' array.
                               Default is None.
        :param list bin_edges: array with numpy histogram style bin_edges.
                               Default is None.
        """
        bin_specs = kwargs.pop('bin_specs', None)
        bin_edges = kwargs.pop('bin_edges', None)

        if bin_specs is not None:
            self.bin_specs = bin_specs
        elif bin_edges is not None:
            if isinstance(bin_edges, np.ndarray):
                bin_edges = bin_edges.tolist()
            assert isinstance(bin_edges, list), 'bin_edges should be a numpy array or list.'
            self.bin_specs = {'bin_edges': bin_edges}
        else:
            self.bin_specs = {}

        # check bin specifications
        if self.bin_specs:
            if not isinstance(self.bin_specs, dict):
                raise TypeError('bin_specs should be a dictionary')
            if not any(k in self.bin_specs for k in ['bin_edges', 'bin_width']):
                raise RuntimeError('unable to interpret bin specifications: {}'.format(self.bin_specs))
            if 'bin_edges' in self.bin_specs and 'bin_width' in self.bin_specs:
                raise RuntimeError('both bin edges and bin width specified for histogram')

    @property
    def bin_specs(self):
        """Get bin_specs dictionary.

        :returns: bin_specs dictionary
        :rtype: dict
        """
        return dict(self._bin_specs)

    @bin_specs.setter
    def bin_specs(self, specs):
        """Set bin-specifications dictionary.

        :param dict bin_specs: dictionary contains 'bin_width' and 'bin_offset' numbers or 'bin_edges' array.
        :raises RunTimeError: If bin_specs already exists, it will not overwrite.
        """
        if hasattr(self, '_bin_specs'):
            raise RuntimeError('histogram bin specifications already set')
        self._bin_specs = dict(specs)

    def value_to_bin_label(self, var_value, greater_equal=False):
        """Return bin index for given bin value.

        :param var_value: variable value for which to find the bin index
        :param bool greater_equal: for float, int, timestamp, return index of bin for which value falls in range
                                   [lower edge, upper edge). If set to true, return index of bin for which value falls
                                   in range [lower edge, upper edge]. Default if false.
        :returns: bin index
        :rtype: int
        """
        # check bin specifications and specified value
        if not self.bin_specs:
            return None
        # NOTE: lines below also work with timestamps. Order is important!
        # find bin label
        if 'bin_width' in self.bin_specs:
            ratio = (var_value - self.bin_specs.get('bin_offset', 0)) / self.bin_specs['bin_width']
            ratio_floor = int(np.floor(ratio))
            if greater_equal and ratio == ratio_floor:
                # correct for upper bin edge
                ratio_floor -= 1
            return ratio_floor
        if var_value <= self.bin_specs['bin_edges'][0]:
            return 0
        elif var_value >= self.bin_specs['bin_edges'][-1]:
            return len(self.bin_specs['bin_edges']) - 2
        for i, be in enumerate(self.bin_specs['bin_edges'][1:]):
            if greater_equal:
                if be >= var_value:
                    return i
            else:
                if be > var_value:
                    return i

    def get_bin_center(self, bin_label):
        """Return bin center for a given bin index.

        :param bin_label: bin label for which to find the bin center
        :returns: bin center, can be float, int, timestamp
        """
        if not self.bin_specs:
            return None
        bin_idx = np.int64(bin_label)
        if 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']
            if bin_idx < 0 or bin_idx >= len(bin_edges):
                raise RuntimeError('bin_label "{}" does not fit in bin edges'.format(bin_label))
            # NOTE: computation below also works with timestamps! Order is
            # important.
            bin_width = bin_edges[bin_idx + 1] - bin_edges[bin_idx]
            bin_width_half = bin_width / 2.
            bin_center = bin_edges[bin_idx] + bin_width_half
        else:
            width = self.bin_specs['bin_width']
            offset = self.bin_specs.get('bin_offset', 0.)
            # NOTE: this notation also works with timestamps!
            bin_center = offset + (bin_idx + 0.5) * width
        return bin_center

    def get_left_bin_edge(self, bin_label):
        """Return left bin edge for a given bin index.

        :param bin_label: bin label for which to find the left bin edge
        :returns: bin edge, can be float, int, timestamp
        """
        # check bin specifications and specified value
        if not self.bin_specs:
            return None
        bin_idx = np.int64(bin_label)
        if 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']
            if bin_idx < 0 or bin_idx >= len(bin_edges):
                raise RuntimeError('bin label "{}" does not fit in bin edges'.format(bin_label))
            bin_edge_left = bin_edges[bin_idx]
        else:
            width = self.bin_specs['bin_width']
            offset = self.bin_specs.get('bin_offset', 0.)
            # NOTE: this notation also works with timestamps!
            bin_edge_left = offset + (bin_idx * width)
        return bin_edge_left

    def get_right_bin_edge(self, bin_label):
        """Return right bin edge for a given bin index.

        :param bin_label: bin label for which to find the right bin edge.
        :returns: bin edge, can be float, int, timestamp
        """
        # check bin specifications and specified value
        if not self.bin_specs:
            return None
        bin_idx = np.int64(bin_label)
        if 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']
            if bin_idx < 0 or bin_idx >= len(bin_edges) - 1:
                raise RuntimeError('bin label "{}" does not fit in bin_edges'.format(bin_label))
            bin_edge_right = bin_edges[bin_idx + 1]
        else:
            width = self.bin_specs['bin_width']
            offset = self.bin_specs.get('bin_offset', 0.)
            # NOTE: this notation also works with timestamps!
            bin_edge_right = offset + (bin_idx + 1) * width
        return bin_edge_right

    def get_bin_edges(self):
        """Return bin edges.

        :returns: bin edges
        :rtype: array
        """
        if not self.bin_specs:
            return None
        bin_edges = self.bin_specs.get('bin_edges', None)
        return bin_edges

    def get_bin_edges_range(self):
        """Return bin range determined from bin edges.

        :returns: bin range
        :rtype: tuple
        """
        if not self.bin_specs:
            return None
        if 'bin_edges' in self.bin_specs:
            return self.bin_specs['bin_edges'][0], self.bin_specs['bin_edges'][-1]
        else:
            return None

    def truncated_bin_edges(self, variable_range=None):
        """Bin edges corresponding to a given variable range.

        :param list variable_range: variable range used for finding the right bin edges array. Optional.
        :returns: truncated bin edges
        :rtype: array
        """
        # check type of input arguments
        if not variable_range:
            return self.get_bin_edges()

        variable_range = tuple(variable_range)

        if not self.bin_specs:
            return np.array(variable_range)

        # check specified value range
        ttest = False
        try:
            if isinstance(variable_range[0], str) or isinstance(variable_range[1], str):
                ttest = True
        except Exception:
            ttest = True
        if ttest or not len(variable_range) == 2 or variable_range[1] <= variable_range[0]:
            raise RuntimeError('expected a variable range (min, max); got "{!s}"'.format(variable_range))
        if 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']
            if variable_range[0] > bin_edges[-1] or variable_range[1] <= bin_edges[0]:
                raise RuntimeError('expected edges in range ({0:f}, {1:f}); got "{2:!s}"'
                                   .format(bin_edges[0], bin_edges[-1], variable_range))

        # on to the non-trivial cases
        if 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']

            min_idx = self.value_to_bin_label(variable_range[0])
            # if variable_range[1] is on a bin-edge, make sure we keep
            # the previous bin's right edge.
            max_idx = self.value_to_bin_label(variable_range[1], greater_equal=True)
            bins = np.asarray(bin_edges[min_idx:max_idx + 2])
            bins = bins.tolist()
        else:
            try:
                bin_width = self.bin_specs['bin_width']
                # bin_offset = self.bin_specs['bin_offset']
            except Exception:
                raise NotImplementedError('no bin_width and/or bin_offset in bin_specs')

            min_idx = self.value_to_bin_label(variable_range[0])
            # if variable_range[1] is on a bin-edge, make sure we keep
            # the previous bin's right edge.
            max_idx = self.value_to_bin_label(variable_range[1], greater_equal=True)
            variable_range = (self.get_left_bin_edge(min_idx), self.get_right_bin_edge(max_idx))

            # ensure to include right edge for highest bin
            bins = np.arange(variable_range[0], variable_range[1] + 0.5 * bin_width, bin_width)
            bins = bins.tolist()

        return bins


class Histogram(BinningUtil, ArgumentsMixin):
    """Generic 1D Histogram class.

    Histogram holds bin labels (name of each bin), value_counts (values of
    the histogram) and a variable name.  The bins can be categoric or
    numeric, where numeric includes timestamps.  In case of numeric bins,
    bin_specs is set.  bins_specs is a dict containing bin_width and
    bin_offset.  In case bins widths are not equal, bin_specs contains
    bin_edges instead of bin_width and bin_offset.
    """

    logger = Logger()

    def __init__(self, counts, **kwargs):
        """Initialize Histogram instance.

        A bin_specs dictionary can be provided as input. bins_specs is a dict
        containing 'bin_width' and 'bin_offset' keys.  In case bins widths are
        not equal, bin_specs contains 'bin_edges' (array) instead of 'bin_width'
        and 'bin_offset'.  'bin_width' and 'bin_offset' can be numeric or numpy
        timestamps.

        Histogram counts can be specified as a ValueCounts object, a dictionary
        or a tuple:

        * tuple: Histogram((bin_values, bin_edges), variable=<your_variable_name>)
        * dict: a dictionary as comes out of pandas.series.value_counts()
          or pandas.Dataframe.groupby.size() over one variable.
        * ValueCounts: a ValueCounts object contains a value_counts dictionary.

        Example bin_specs dictionaries are:

        >>> bin_specs = { 'bin_width': 1, 'bin_offset': 0 }
        >>> bin_spect = { 'bin_edges': [0,2,3,4,5,7,8] }
        >>> bin_specs = { 'bin_width': np.timedelta64(30,'D'),
                          'bin_offset': np.datetime64('2010-01-04') }

        :param counts: histogram counts
        :param dict bin_specs: dictionary contains 'bin_width' and 'bin_offset' numbers or 'bin_edges' array
                               (default is None)
        :param str variable: name of the variable represented by the histogram
        :param type datatype: data type of the variable represented by the histogram (optional)
        """
        # initialize Binning, pass bin_specs from kwargs
        BinningUtil.__init__(self, bin_specs=kwargs.pop('bin_specs', None))

        # parse arguments
        self._process_kwargs(kwargs, variable='', datatype=None)
        self.check_arg_vals('variable')
        self.check_arg_types(variable=str, bin_specs=dict)

        # check for extraneous keyword arguments
        self.check_extra_kwargs(kwargs)

        self.logger.debug('Initializing histogram for "{variable}".', variable=self.variable)

        # set value counts
        if isinstance(counts, ValueCounts):
            # from ValueCounts object
            self._from_value_counts(counts)
        elif isinstance(counts, dict):
            # from value-counts dictionary
            self._from_dict(counts)
        else:
            # try from NumPy-style histogram
            try:
                counts, var_vals = list(counts[0]), list(counts[1])
            except Exception:
                self.logger.fatal('Invalid type for specified counts: "{type}".', type=type(counts).__name__)
                raise RuntimeError('Invalid type for specified counts.')
            self._from_numpy(counts, var_vals)

        # check counts
        if self._val_counts.num_nonone_bins < 1:
            self.logger.warning('No bin counts specified for "{variable}".', variable=self.variable)

        # check bin specifications
        if self.bin_specs:
            if 'bin_width' in self.bin_specs:
                try:
                    _check_num_vals(self.get_bin_labels())
                except TypeError as exc:
                    self.logger.fatal(
                        'Non-numeric values in bin labels not allowed if bin width is specified: {labels}.',
                        labels=self.get_bin_labels())
                    raise exc
            elif 'bin_edges' in self.bin_specs:
                edges = self.bin_specs['bin_edges']
                if len(edges) != self._val_counts.num_nonone_bins + 1:
                    self.logger.fatal(
                        'Number of specified bin edges ({n_edges:d}) does not match number of bins ({n_bins:d}).',
                        n_edges=len(edges), n_bins=self._val_counts.num_nonone_bins)
                    raise RuntimeError('numbers of bins and bin edges do not match')
                try:
                    _check_num_vals(edges)
                except TypeError as exc:
                    self.logger.fatal('Non-numeric values found in bin edges: {edges}.', edges=edges)
                    raise exc
                if any(e2 <= e1 for e1, e2 in zip(edges[:-1], edges[1:])):
                    self.logger.fatal('Values of bin edges are not increasing: {edges!s}', edges=edges)
                    raise RuntimeError('Invalid bin edges specified.')

    @property
    def variable(self):
        """Name of variable represented by the histogram.

        :returns: variable name
        :rtype: string
        """
        return self._variable

    @variable.setter
    def variable(self, var):
        """Set name of variable represented by the histogram.

        :param str var: name of variable represented by the histogram.
        :raises RunTimeError: If a variable name already exists, it will not overwritten.
        """
        if hasattr(self, '_variable'):
            raise RuntimeError('histogram variable already set')
        self._variable = str(var)

    @property
    def datatype(self):  # noqa
        """Data type of the variable represented by the histogram.

        :returns: data type
        :rtype: type
        """
        return self._datatype

    @datatype.setter
    def datatype(self, dt):
        """Set data type of the variable represented by the histogram.

        :param type dt: type of the variable represented by the histogram
        :raises RunTimeError: If datatype has already been set, it will not overwritten
        """
        if hasattr(self, '_datatype'):
            raise RuntimeError('datatype already set')
        self._datatype = dt

    @property
    def n_dim(self):  # noqa
        """Number of histogram dimensions.

        The number of histogram dimensions, which is equal to one by
        construction.

        :returns: number of dimensions
        :rtype: int
        """
        return 1

    @property
    def n_bins(self):  # noqa
        """Number of bins in the ValueCounts object.

        :returns: number of bins
        :rtype: int
        """
        return self._val_counts.num_bins

    @property
    def num_bins(self):  # noqa
        """Number of bins.

        :returns: number of bins
        :rtype: int
        """
        return self._val_counts.num_bins

    def get_bin_labels(self):
        """Return all bin labels.

        :returns: array of all bin labels
        :rtype: array
        """
        return [v[0] for v in self._val_counts.nononecounts.keys()]

    def bin_labels(self):
        """Return bin labels.

        :returns: array of all bin labels
        :rtype: array
        """
        return np.array(self.get_bin_labels())

    def get_uniform_bin_edges(self):
        """Return numpy style bin-edges array with uniform binning.

        :returns: array of all bin edges
        :rtype: array
        """
        if not self.bin_specs:
            return None
        bin_range = list(self.get_nonone_bin_range())
        return self.truncated_bin_edges(variable_range=bin_range)

    def get_nonone_bin_edges(self):
        """Return numpy style bin-edges array.

        :returns: array of the bin edges
        :rtype: array
        """
        if not self.bin_specs:
            return None
        if 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']
        else:
            width = self.bin_specs['bin_width']
            offset = self.bin_specs.get('bin_offset', 0)
            edges = [offset + v[0] * width for v in self._val_counts.nononecounts.keys()]
            bin_edges = edges + [edges[-1] + width]
        return bin_edges

    def bin_edges(self):
        """Return numpy style bin_edges array with uniform binning.

        :returns: array of all bin edges
        :rtype: array
        """
        return np.array(self.get_uniform_bin_edges())

    def get_nonone_bin_centers(self):
        """Return bin centers.

        :returns: array of the bin centers
        :rtype: array
        """
        if not self.bin_specs or len(self.bin_specs) == 0:
            return self.get_bin_labels()
        if 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']
            # NOTE: computation below also works with timestamps! Order is
            # important.
            bin_centers = []
            for i in range(len(bin_edges) - 1):
                bin_width = bin_edges[i + 1] - bin_edges[i]
                bin_width_half = bin_width / 2.
                bin_center = bin_edges[i] + bin_width_half
                bin_centers.append(bin_center)
        else:
            width = self.bin_specs['bin_width']
            offset = self.bin_specs.get('bin_offset', 0)
            bin_centers = [offset + (v[0] + 0.5) * width for v in self._val_counts.nononecounts.keys()]
        return bin_centers

    def bin_centers(self):
        """Return bin centers.

        :returns: array of the bin centers
        :rtype: array
        """
        return np.array(self.get_nonone_bin_centers())

    def get_bin_count(self, bin_label):
        """Get bin count for specific bin label.

        :param bin_label: a specific key to find corresponding bin.
        :returns: bin counter value
        :rtype: int
        """
        return self._val_counts.nononecounts.get((bin_label,), 0)

    def get_nonone_bin_counts(self):
        """Return bin counts.

        :returns: array of the bin counts
        :rtype: array
        """
        if not self.bin_specs or len(self.bin_specs) == 0:
            bin_counts = self._val_counts.nononecounts.values()
        elif 'bin_edges' in self.bin_specs:
            bin_edges = self.bin_specs['bin_edges']
            # NOTE: computation below also works with timestamps! Order is
            # important.
            bin_centers = []
            for i in range(len(bin_edges) - 1):
                bin_width = bin_edges[i + 1] - bin_edges[i]
                bin_width_half = bin_width / 2.
                bin_center = bin_edges[i] + bin_width_half
                bin_centers.append(bin_center)
            bin_counts = [self.get_hist_val(bc) for bc in bin_centers]
        else:
            bin_counts = self._val_counts.nononecounts.values()
        return bin_counts

    def bin_entries(self):
        """Return number of bin entries.

        Return the bin counts of the known bins in the value_counts object.

        :returns: array of the bin counts
        :rtype: array
        """
        return np.array(self.get_nonone_bin_counts())

    def get_bin_range(self):
        """Return the bin range.

        :returns: tuple of the bin range found
        :rtype: tuple
        """
        return self.get_nonone_bin_range()

    def get_nonone_bin_range(self):
        """Return the bin range.

        :returns: tuple of the bin range found
        :rtype: tuple
        """
        if not self.bin_specs:
            return None
        if 'bin_edges' in self.bin_specs:
            return self.bin_specs['bin_edges'][0], self.bin_specs['bin_edges'][-1]
        vals = [v[0] for v in self._val_counts.nononecounts]
        min_max = (vals[0], vals[-1])
        if self.bin_specs:
            width = self.bin_specs['bin_width']
            offset = self.bin_specs.get('bin_offset', 0.)
            return tuple(offset + m * width for m in (min_max[0], min_max[1] + 1))
        return min_max

    def get_hist_val(self, var_value):
        """Get bin count for bin by value of histogram variable.

        :param var_value: a specific value to find corresponding bin.
        :returns: bin counter value
        :rtype: int
        """
        try:
            bin_label = self.value_to_bin_label(var_value)
        except Exception as exc:
            self.logger.error('bin label for variable value "{value!s}" not found ({error})',
                              value=var_value, error=exc)
            return 0
        return self.get_bin_count(bin_label)

    def get_bin_vals(self, variable_range=None, combine_values=True):
        """Get bin labels/edges and corresponding bin counts.

        Bin values corresponding to a given variable range.

        :param list variable_range: variable range used for finding the right bins to get values from. Optional.
        :param bool combine_values: if bin_specs is not set, combine existing bin labels with variable range.
        :returns: two arrays of bin values and bin edges
        :rtype: array
        """
        # check type of input arguments
        if variable_range is not None:
            variable_range = tuple(variable_range)
        combine_values = bool(combine_values)

        # create NumPy arrays of bin labels/edges
        if variable_range and not self.bin_specs:
            labs = bins = np.unique(np.append(self.get_bin_labels(), variable_range)) if combine_values else \
                np.array(variable_range)
        else:
            labs = np.array(self.get_bin_labels())
            bins = labs if not self.bin_specs else np.array(
                self.get_nonone_bin_edges())

        if variable_range and self.bin_specs:
            bins = self.truncated_bin_edges(variable_range=list(variable_range))
            # NOTE: computation below also works with timestamps! Order is
            # important.
            bin_centers = []
            for i in range(len(bins) - 1):
                bin_width = bins[i + 1] - bins[i]
                bin_width_half = bin_width / 2.
                bin_center = bins[i] + bin_width_half
                bin_centers.append(bin_center)
            labs = [self.value_to_bin_label(bc) for bc in bin_centers]

        return np.array([self.get_bin_count(v) for v in labs]), np.asarray(bins)

    def remove_keys_of_inconsistent_type(self, prefered_key_type=None):
        """Remove all keys that have inconsistent data type(s).

        :param tuple prefered_key_type: the prefered key type to keep. Can be a tuple, list, or single type. E.g. str
                                        or (int,str,float). If None provided, the most common key type found is kept.
        """
        n_keys_prev = len(self._val_counts._cnts)
        self._val_counts.remove_keys_of_inconsistent_type(prefered_key_type)
        n_keys_new = len(self._val_counts._cnts)

        if n_keys_new < n_keys_prev:
            self.logger.info('Removed "{n:d}" inconsistent keys out of "{n_total:d}", requiring "{type}" data type.',
                             n=n_keys_prev - n_keys_new, n_total=n_keys_prev,
                             type=prefered_key_type if prefered_key_type is not None else 'most common')

    def simulate(self, size, *args):
        """Simulate data using self (Histogram instance) as PDF.

        see https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_continuous.html

        :param int size: number of data points to generate
        :return numpy.array generated_data: the generated data
        :returns: Histogram of the generated data
        :rtype: Histogram
        """
        h_norm = self.to_normalized()
        hist, bins = h_norm.get_bin_vals(*args)
        values = np.random.rand(size)

        # check if Histogram is categorical or numeric
        if not self.bin_specs:
            generated_data = np.random.choice(bins, p=hist, size=size)
            vc = pd.Series(generated_data).value_counts() / float(len(generated_data))
            h_sim = Histogram((vc.values, vc.index.values), variable='x')
        else:
            # see
            # http://stackoverflow.com/questions/17821458/random-number-from-histogram
            bin_midpoints = bins[:-1] + np.diff(bins) / 2
            cdf = np.cumsum(hist)
            generated_data = np.searchsorted(cdf, values)
            random_from_cdf = bin_midpoints[generated_data]
            h_sim = Histogram(np.histogram(random_from_cdf, bins=bins), variable=self.variable)
        return generated_data, h_sim

    def surface(self):
        """Calculate surface of the histogram.

        :returns: surface
        """
        if not self.bin_specs:
            raise NotImplementedError(
                'Surface for categorical histograms is not implemented.')
        hist, bin_edges = self.get_bin_vals()
        bin_widths = np.diff(bin_edges)
        return np.multiply(hist, bin_widths).sum()

    def copy(self, **kwargs):
        """Return a copy of this histogram.

        :param str variable: assign new variable name
        """
        new_var_name = str(kwargs.pop('variable', self.variable))
        return Histogram(counts=self.get_bin_vals(**kwargs), variable=new_var_name)

    def to_normalized(self, **kwargs):
        """Return a normalized copy of this histogram.

        :param str new_var_name: assign new variable name
        :param list variable_range: variable range used for finding the right bins to get values from.
        :param bool combine_values: if bin_specs is not set, combine existing bin labels with variable range.
        """
        # convert to normalized histogram
        new_var_name = str(kwargs.pop('variable', self.variable))
        bin_vals = self.get_bin_vals(**kwargs)
        values = np.float64(bin_vals[0]) / bin_vals[0].sum()
        # When values is a numpy array of 1 element np.float64() returns a 0-dimensional array. See
        # https://github.com/numpy/numpy/issues/3161. The following
        # if-statement is a workaround for this issue.
        if not values.shape:
            values = values.reshape((1,))
        return Histogram(counts=(values, bin_vals[1]), variable=new_var_name)

    @classmethod
    def combine_hists(cls, hists, labels=False, rel_bin_width_tol=1e-6, **kwargs):
        """Combine a set of histograms.

        :param array hists: array of Histograms to add up.
        :param label labels: histograms to add up have labels? (else are numeric) Default is False.
        :param str variable: name of variable described by the summed-up histogram
        :param float rel_bin_width_tol: relative tolerance between numeric bin edges.
        :returns: summed up histogram
        :rtype: Histogram
        """
        # parse input arguments
        new_var_name = str(kwargs.pop('variable', 'x'))
        cls.logger.debug('Combining histograms into one new histogram for "{name}".', name=new_var_name)

        try:
            bin_vals = [h.get_bin_vals(**kwargs) for h in hists]
        except Exception:
            cls.logger.fatal('Unable to get bin values from specified histograms ({hists!s}).', hists=hists)
            raise RuntimeError('Invalid input histograms specified.')
        if not bin_vals:
            raise RuntimeError('No input histograms specified.')

        # check histogram bins
        ref_bins = bin_vals[0][1]
        if labels:
            if not all(np.array_equal(ref_bins, h[1]) for h in bin_vals):
                cls.logger.fatal('Histograms with different binnings specified: {hists}.',
                                 hists=', '.join(str(h[1]) for h in hists))
                raise RuntimeError('histograms with different binnings specified')
        else:
            ref_bin_width = hists[0].bin_specs['bin_width']
            atol = rel_bin_width_tol * ref_bin_width
            if not all(np.allclose(ref_bins, h[1], rtol=0, atol=atol) for h in bin_vals):
                cls.logger.fatal('Histograms with different binnings specified: {hists}.',
                                 hists=', '.join(str(h[1]) for h in hists))
                raise RuntimeError('histograms with different binnings specified')

        # return histogram with sum of bin counts
        return cls(counts=(np.sum((h[0] for h in bin_vals), axis=0), ref_bins), variable=new_var_name)

    def _from_value_counts(self, counts):
        """Create Histogram from ValueCounts instance.

        :param ValueCounts counts: value counts
        """
        # initialize value counts from ValueCounts object
        if self.variable not in counts.key:
            self.logger.fatal('Variable "{var}" not in value counts with variables ({vars})',
                              var=self.variable, vars=', '.join('"{}"'.format(v) for v in counts.key))
            raise RuntimeError('specified variable and counts do not match')
        self._val_counts = counts.create_sub_counts((self.variable,))

    def _from_dict(self, counts):
        """Create Histogram from dictionary.

        :param dict counts: value counts
        """
        # initialize value counts from dictionary
        counts = dict((k if isinstance(k, tuple) else (k,), v) for k, v in counts.items())
        if not all(len(k) == 1 for k in counts):
            self.logger.fatal('Specified counts dictionary contains keys with multiple variable values.')
            raise AssertionError('Invalid format for counts keys.')
        _check_num_vals(iter(counts.values()))
        self._val_counts = ValueCounts((self.variable,), (self.variable,), counts)

    def _from_numpy(self, counts, bin_edges):
        """Create Histogram from NumPy-style histogram.

        :param array counts: numpy histogram counts array
        :param array bin_edges: bin edges
        """
        # initialize from NumPy-style histogram
        _check_num_vals(counts)
        if len(counts) == len(bin_edges) - 1:
            # interpret specified variable values as bin edges
            del self._bin_specs
            self.bin_specs = {'bin_edges': list(bin_edges)}
            bin_edges = list(range(len(counts)))
        elif len(counts) != len(bin_edges):
            # cannot interpret specified variable values as bin values
            self.logger.fatal('Numbers of specified variable values ({n:d}) and value counts ({counts:d}) do not match',
                              n=len(bin_edges), counts=len(counts))
            raise AssertionError('specified variable values and value counts do not match')
        self._val_counts = ValueCounts((self.variable,), (self.variable,),
                                       dict(((v,), c) for c, v in zip(counts, bin_edges)))



from eskapade.core_ops.links.assert_in_ds import AssertInDs
from eskapade.core_ops.links.break_link import Break
from eskapade.core_ops.links.ds_object_deleter import DsObjectDeleter
from eskapade.core_ops.links.ds_to_ds import DsToDs
from eskapade.core_ops.links.event_looper import EventLooper
from eskapade.core_ops.links.hello_world import HelloWorld
from eskapade.core_ops.links.ipython_embed import IPythonEmbed
from eskapade.core_ops.links.line_printer import LinePrinter
from eskapade.core_ops.links.print_ds import PrintDs
from eskapade.core_ops.links.repeat_chain import RepeatChain
from eskapade.core_ops.links.skip_chain_if_empty import SkipChainIfEmpty
from eskapade.core_ops.links.to_ds_dict import ToDsDict
from eskapade.core_ops.links.apply import DsApply
from eskapade.core_ops.links.import_data_store import ImportDataStore

__all__ = ['AssertInDs',
           'Break',
           'DsObjectDeleter',
           'DsToDs',
           'EventLooper',
           'HelloWorld',
           'IPythonEmbed',
           'LinePrinter',
           'PrintDs',
           'RepeatChain',
           'SkipChainIfEmpty',
           'ToDsDict',
           'DsApply',
           'ImportDataStore']



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk207_record_vectorizer

Created: 2017/03/13

Description:
    This macro performs the vectorization of an input column of an input dataframe.
    E.g. a columnn x with values 1, 2 is tranformed into columns x_1 and x_2,
    with values True or False assigned per record.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, resources, Chain
from eskapade import core_ops, analysis
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk207_record_vectorizer')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk207_record_vectorizer'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# --- Set path of data
data_path = resources.fixture('dummy.csv')

#########################################################################################
# --- now set up the chains and links based on configuration flags

ch1 = Chain('MyChain1')

# --- read dummy dataset
read_data = analysis.ReadToDf(key='test1', sep='|', reader='csv', path=data_path)
ch1.add(read_data)

# --- print contents of the datastore
pds = core_ops.PrintDs(name='printer1')
pds.keys = ['test1']
ch1.add(pds)

# --- add the record vectorizer
#     Here the columns x and y of the input dataset are vectorized
#     e.g. x=1 becomes the column: x_1 = True
vectorizer = analysis.RecordVectorizer()
vectorizer.columns = ['x', 'y']
vectorizer.read_key = 'test1'
vectorizer.store_key = 'vect_test'
vectorizer.astype = int
ch1.add(vectorizer)

# --- print contents of the datastore
pds = core_ops.PrintDs(name='printer2')
pds.keys = ['vect_test']
ch1.add(pds)

#########################################################################################

logger.debug('Done parsing configuration file esk207_record_vectorizer')



"""Project: Eskapade - A python-based package for data analysis.

Class: EventLooper

Created: 2016/11/08

Description:
    EventLooper algorithm processes input lines and reprints them,
    e.g. to use with map/reduce

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy
import sys

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class EventLooper(Link):
    """Event looper algorithm processes input lines and reprints or stores them.

    Input lines are taken from sys.stdin, processed, and printed on screen.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str filename: file name where the strings are located (txt or similar). Default is None. (optional)
        :param str store_key: key to collect in datastore. If set lines are collected. (optional)
        :param list line_processor_set: list of functions to apply to input lines. (optional)
        :param bool sort: if true, sort lines before storage (optional)
        :param bool unique: if true, keep only unique lines before storage (optional),
        :param list skip_line_beginning_with: skip line if it starts with any of the list. input is list of strings.
            Default is ['#'] (optional)
        """
        # initialize Link
        Link.__init__(self, kwargs.pop('name', 'EventLooper'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             filename=None,
                             store_key=None,
                             line_processor_set=[],
                             sort=False,
                             unique=False,
                             skip_line_beginning_with=['#'])

        # process keyword arguments
        self.check_extra_kwargs(kwargs)

        # default line stream to pick up lines is set to sys.stdin below
        # input stream and possible input file
        self._f = None
        self._linestream = None

        # collect lines for storage
        self._collect = False

    def initialize(self):
        """Perform basic checks of configured attributes."""
        if self.store_key is not None:
            assert isinstance(self.store_key, str) and self.store_key, 'output key not set.'
            self._collect = True

        # default line stream is set to sys.stdin
        self._linestream = sys.stdin

        # try to open input file, if provided.
        # if successful, switch linestream to file.
        if self.filename is not None:
            assert isinstance(self.filename, str) and len(self.filename), 'input file name not set properly.'
            try:
                self._f = open(self.filename, "r")
            except IOError:
                raise Exception('Cannot open file {}.'.format(self.filename))
            # successful, so switch linestream to file.
            self._linestream = self._f

        return StatusCode.Success

    def execute(self):
        """Process all incoming lines.

        No output is printed except for lines that are passed on,
        such that the output lines can be picked up again by another parser.
        """
        lines = []

        # default line stream is set to sys.stdin
        # print or collect (processed) lines
        for line in self._linestream:
            line = line.strip()
            # skip empty and comment lines
            if not line or any(line.startswith(c) for c in self.skip_line_beginning_with):
                continue
            myline = copy.deepcopy(line)
            for func in self.line_processor_set:
                myline = func(myline)
            if not self._collect:
                print(myline)
            else:
                lines.append(myline)

        if not self._collect:
            return StatusCode.Success

        # perform basic operations before storage, if desired:
        # sorting and unique set.
        if self.sort:
            lines = sorted(lines)
        if self.unique:
            lines = list(set(lines))

        ds = process_manager.service(DataStore)
        ds[self.store_key] = lines
        ds['n_' + self.store_key] = len(lines)

        return StatusCode.Success

    def finalize(self):
        """Close open file if present."""
        if self._f is not None:
            self._f.close()

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Class: FixPandasDataFrame

Created: 2017/04/07

Description:
    Link for fixing dirty pandas dataframe with inconsistent datatypes
    See example in: tutorials/esk501_fix_pandas_dataframe.py

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy
import re
import string
from collections import Counter

import numpy as np
import pandas as pd
import fastnumbers

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager
from eskapade.data_quality.dq_helper import bool_to_int, check_nan, convert, cleanup_string, CONV_FUNCS


class FixPandasDataFrame(Link):
    """Fix dirty Pandas dataframe with inconsistent datatypes.

    Default settings perform the following clean-up steps on an input
    dataframe:

    - Fix all column names.  E.g. remove punctuation and strange characters,
      and convert spaces to underscores.
    - Check for various possible nans in the dataset, then make all nans
      consistent by turning them into numpy.nan (= float)
    - Per column, assess dynamically the most consistent datatype (ignoring
      all nans in that column).  E.g. bool, int, float, datetime64, string.
    - Per column, make the data types of all rows consistent, by using the
      identified (or imposed) data type (by default ignoring all nans)

    Boolean columns with contamination get converted to string columns by
    default.  Optionally, they can be converted to integer columns as well.

    The FixPandasDataFrame link can be used in a dataframe loop, in which
    case any data type assessed per column in the first dataframe iteration
    will be used for the next dataframes as well.

    The default settings should work pretty well in many circumstances, by
    can be configured pretty flexibly.  Optionally:

    - Instead of dynamically assessed, the data type can also be imposed per column
    - All nans in a column can be converted to a value consistent with the
      data type of that column.  E.g. for integer columns, nan -> -999
    - An alternative nan can be set per column and datatype
    - Modifications can be applied inplace, i.e. directly to the input dataframe
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str read_key: key of input data to read from data store
        :param bool copy_columns_from_df: if true, copy all columns from the dataframe (default is true)
        :param list original_columns: original (unfixed) column names to pick up from input data
                                      (required if copy_columns_from_df is set to false)
        :param list contaminated_columns: (original) columns that are known to have mistakes and that
                                          should be fixed (optional)
        :param bool fix_column_names: if true, fix column names (default is true)
        :param bool strip_hive_prefix: if true, strip table-name (hive) prefix from column names,
                                       e.g. table.bla -> bla (default is false)
        :param bool convert_inconsistent_dtypes: fix column datatypes in case of data type inconsistencies in rows
                                                 (default is true)
        :param dict var_dtype: dict forcing columns to certain datatypes, e.g. {'A': int} (optional)
        :param dict var_convert_inconsistent_dtypes: dict allowing one to overwrite if certain columns
                                                     datatypes should be fixed, e.g. {'A': False} (optional)
        :param dict var_convert_func: dict with datatype conversion functions for certain columns
        :param check_nan_func: boolean return function to check for nans in columns.
                               (default is None, in which case a standard checker function gets picked up)
        :param bool convert_inconsistent_nans: if true, convert all nans to data type consistent with rest of column
                                               (default is false)
        :param dict var_convert_inconsistent_nans: dict allowing one to overwrite if certain column
                                                   nans should be fixed, e.g. {'A': False} (optional)
        :param dict var_nan: dict with nans for certain columns (optional)
        :param dict nan_dtype_map: dictionary of nans for given data types, e.g. { int: -999 }
        :param nan_default: default nan value to which all nans found get converted (default is numpy.nan)
        :param list var_bool_to_int: convert boolean column to int (default is conversion of boolean to string)
        :param bool inplace: replace original columns; overwrites store_key to read_key (default is False)
        :param str store_key: key of output data to store in data store
        :param bool drop_dup_rec: if true, drop duplicate records from data frame after other fixes (default is false)
        :param bool strip_string_columns: if true, apply strip command to string columns (default is true)
        :param list cleanup_string_columns: boolean or list. apply cleaning-up to list of selected or
            all string columns. More aggressive than strip. Default is empty (= false).
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'FixPandasDataFrame'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             read_key='',
                             copy_columns_from_df=True,
                             original_columns=[],
                             contaminated_columns=[],
                             fix_column_names=True,
                             strip_hive_prefix=False,
                             convert_inconsistent_dtypes=True,
                             check_nan_func=None,
                             convert_inconsistent_nans=False,
                             var_convert_inconsistent_dtypes={},
                             var_convert_func={},
                             var_convert_inconsistent_nans={},
                             var_nan={},
                             var_dtype={},
                             var_bool_to_int=[],
                             inplace=False,
                             store_key='',
                             nan_dtype_map={},
                             nan_default=np.nan,
                             drop_dup_rec=False,
                             strip_string_columns=True,
                             cleanup_string_columns=[])

        # check residual kwargs. exit if any present.
        self.check_extra_kwargs(kwargs)

        self.fixed_columns = []

        self._var_dicts = [self.var_dtype, self.var_convert_inconsistent_dtypes,
                           self.var_nan, self.var_convert_func, self.var_convert_inconsistent_nans]
        self._df_orig_dtype = {}
        self._cnts = {}

        # set nans for individual data types
        if np.int64 not in self.nan_dtype_map:
            self.nan_dtype_map[np.int64] = -999
        if np.float64 not in self.nan_dtype_map:
            self.nan_dtype_map[np.float64] = np.nan
        if np.datetime64 not in self.nan_dtype_map:
            self.nan_dtype_map[np.datetime64] = np.datetime64('NaT')
        if str not in self.nan_dtype_map:
            self.nan_dtype_map[str] = 'not_a_str'
        if bool not in self.nan_dtype_map:
            self.nan_dtype_map[bool] = 'not_a_bool'
        if np.bool_ not in self.nan_dtype_map:
            self.nan_dtype_map[np.bool_] = self.nan_dtype_map[bool]

    def initialize(self):
        """Initialize the link."""
        self.check_arg_types(read_key=str, store_key=str)
        self.check_arg_types(recurse=True, allow_none=True, original_columns=str)
        self.check_arg_vals('read_key')

        if not isinstance(self.cleanup_string_columns, list) and not isinstance(self.cleanup_string_columns, bool):
            raise AssertionError('cleanup_string_columns should be a list of column names or boolean.')

        if self.read_key == self.store_key:
            self.inplace = True
            self.logger.debug('store_key equals read_key; inplace has been set to "True".')

        if self.inplace:
            self.store_key = self.read_key
            self.logger.debug('store_key has been set to read_key "{key}".', key=self.store_key)

        if not self.store_key:
            self.store_key = self.read_key + '_fix'
            self.logger.debug('store_key has been set to "{key}".', key=self.store_key)

        # check data types
        for k in self.var_dtype:
            if k not in self.contaminated_columns:
                self.contaminated_columns.append(k)
            try:
                # convert to consistent types
                dt = np.dtype(self.var_dtype[k]).type
                if dt is np.str_ or dt is np.object_:
                    dt = str
                self.var_dtype[k] = dt
            except BaseException:
                raise TypeError('Unknown assigned datatype to variable "{}".'.format(k))

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Fixing the Pandas dataframe consists of four steps:

        - Fix all column names. E.g. remove punctuation and strange characters, and convert spaces to underscores.
        - Check existing nans in that dataset, and make all nans consistent, for easy conversion later on.
        - Assess most consistent datatype for each column (ignoring all nans)
        - Make data types in each row consistent (by default ignoring all nans)
        """
        ds = process_manager.service(DataStore)

        # basic checks on contensts of the data frame
        if self.read_key not in ds:
            raise KeyError('key "{}" not in DataStore.'.format(self.read_key))
        df = ds[self.read_key]
        if not isinstance(df, pd.DataFrame):
            raise TypeError('Retrieved object not of type pandas DataFrame.')
        if len(df.index) == 0:
            raise AssertionError('Dataframe "{}" is empty.'.format(self.read_key))

        # use all columns from the dataframe?
        if self.copy_columns_from_df and not self.original_columns:
            self.original_columns = df.columns.tolist()
        # check not empty
        if not self.original_columns:
            self.logger.warning('Original columns have not been set; nothing to do.')
            return StatusCode.Recoverable

        self.logger.debug('Original columns to be processed:\n{cols!s}.', cols=self.original_columns)

        # check presence and data types of requested columns
        for col in self.original_columns:
            if col not in df.columns:
                raise AssertionError('Column "{}" not present in input data frame {}.'.format(col, self.read_key))

        # set string columns to clean up
        if isinstance(self.cleanup_string_columns, bool):
            self.cleanup_string_columns = copy.copy(self.original_columns) if self.cleanup_string_columns else []

        # before we starting fixing, make a copy of the dataframe
        df_ = df if self.inplace else df.copy()

        # 1. fix column names
        self.fixed_columns = copy.copy(self.original_columns)
        if self.fix_column_names:
            regex = re.compile('[{}]'.format(re.escape(string.punctuation)))
            all_columns = df_.columns.tolist()
            for i, col in enumerate(all_columns[:]):
                if col not in self.original_columns:
                    continue
                # strip hive prefix. eg 'tab1.foo' -> 'foo'
                new_col = col.strip()
                if self.strip_hive_prefix:
                    carr = col.split('.')
                    if len(carr) >= 2:
                        # remove prefix
                        carr.pop(0)
                    new_col = '.'.join(carr)
                # replace empty spaces
                new_col = new_col.replace(' ', '_')
                # replace punctuation
                new_col = regex.sub('_', new_col)
                # keep only alphanumeric and _
                new_col = re.sub('[^A-Za-z0-9_]+', '', new_col)
                if not new_col:
                    raise ValueError('Column name "{}" empty after cleaning.'.format(col))
                # replace column names
                all_columns[i] = new_col
                self.fixed_columns[self.fixed_columns.index(col)] = new_col
                if col in self.cleanup_string_columns:
                    self.cleanup_string_columns[self.cleanup_string_columns.index(col)] = new_col
                if col in self.var_bool_to_int:
                    self.var_bool_to_int[self.var_bool_to_int.index(col)] = new_col
                if col in self.contaminated_columns:
                    self.contaminated_columns[self.contaminated_columns.index(col)] = new_col
                for vd in self._var_dicts:
                    if col in vd:
                        vd[new_col] = vd.pop(col)
            df_.columns = all_columns
            self.logger.debug('Fixed column names are:\n{cols!s}.', cols=self.fixed_columns)

        # --- Next: fix datatypes - all rows in a column get consistent datatype, except for nans

        # init - data types not object nor float have already been assessed correctly by pandas
        #        and have no polution. Can skip these.
        for col in set(self.fixed_columns).difference(self.var_dtype):
            # convert to consistent types
            dt = df_[col].dtype.type
            if dt is not np.object_ and dt is not np.float64:
                self.var_dtype[col] = dt

        # init - assess data types of columns as earlier assessed by pandas
        for col in set(self.fixed_columns).difference(self._df_orig_dtype):
            # convert to consistent types
            dt = df_[col].dtype.type
            if dt is np.str_ or dt is np.object_:
                dt = str
            self._df_orig_dtype[col] = dt

        # 2. make all nans consistent, for easy conversion later on
        #    check on nans
        n_df = len(df_.index)
        is_nan = pd.DataFrame(index=df_.index)
        check_nan_func = check_nan if not self.check_nan_func else self.check_nan_func
        for col in self.fixed_columns:
            is_nan[col] = df_[col].apply(check_nan_func)
            n_nan = is_nan[col].values.sum()
            self._cnts[col] = Counter({'NaN': n_nan})
            if n_nan:
                self.logger.debug('Column "{col}" contains {n:d} NaNs out of {total:d}.', col=col, n=n_nan, total=n_df)
            dt = self._df_orig_dtype[col]
            # np.nan is a float, so for non-floats nan is considered contamination by pandas
            if n_nan and dt is not np.float64:
                if col not in self.contaminated_columns:
                    self.contaminated_columns.append(col)
        df_[is_nan.columns] = df_[is_nan.columns].mask(is_nan, self.nan_default)

        # 3. multiple datatypes in columns besides nans?
        #    find most common one per column
        keep = ~is_nan
        for col in set(self.fixed_columns).difference(self.var_dtype):
            try:
                # fast, but does not work for timestamps
                df_[col] = df_[col].apply(fastnumbers.fast_real)
            except:
                # slow, but works for timestamps
                df_[col] = df_[col].apply(convert)
            dfcol = df_[col][keep[col]]
            dtype_cnt = Counter(dfcol.apply(type).value_counts().to_dict())
            # for bookkeeping
            self._cnts[col].update(dtype_cnt)
            if len(dtype_cnt) == 0:
                # column contains only nans!
                if col not in self.var_dtype:
                    # stick to dtype as determined by pandas (= float)
                    self.var_dtype[col] = self._df_orig_dtype[col]
                continue
            # store most common datatype
            # first convert to consistent types
            for dtp in dtype_cnt:
                ndt = np.dtype(dtp).type
                if ndt is np.str_ or ndt is np.object_:
                    ndt = str
                mc = dtype_cnt.pop(dtp)
                dtype_cnt[ndt] += mc
            preferred_dtype = determine_preferred_dtype(dtype_cnt)
            if col not in self.var_dtype:
                self.var_dtype[col] = preferred_dtype
            if len(dtype_cnt) > 1:
                self.logger.warning('Found multiple types for column "{col!s}"', col=col)
                self.logger.debug('Picked type "{type}" for column "{col}" (counts: {counts!s}).',
                                  type=preferred_dtype, col=col, counts=dtype_cnt)
                if col not in self.contaminated_columns:
                    self.contaminated_columns.append(col)

        self.logger.debug('Consider setting link.var_dtype = {type!s}', type=self.var_dtype)
        self.logger.info('Fixing contamination in columns {cols}',
                         cols=', '.join('"{}"'.format(c) for c in self.contaminated_columns))

        # 4. fix contamination in each column
        for col in self.contaminated_columns:
            dt = self.var_dtype[col]
            self.logger.debug('Converting rows in column "{col}" to type "{type}".', col=col, type=dt)
            # pick conversion function of choice
            if col in self.var_convert_func:
                fnc = self.var_convert_func[col]
            elif col in self.var_bool_to_int:
                fnc = bool_to_int
            elif dt in CONV_FUNCS:
                fnc = CONV_FUNCS[dt]
            else:
                raise RuntimeError('Do not know how to convert column "{}"'.format(col))
            # convert inconsistent dtypes?
            convert_inconsistent_dtypes = self.var_convert_inconsistent_dtypes.get(
                col, self.convert_inconsistent_dtypes)
            convert_inconsistent_nans = self.var_convert_inconsistent_nans.get(col, self.convert_inconsistent_nans)
            fnc_kw = {'convert_inconsistent_dtypes': convert_inconsistent_dtypes}
            # pick nan of choice
            if col in self.var_nan:
                fnc_kw['nan'] = self.var_nan[col]
                if convert_inconsistent_nans:
                    rdt = dt if col not in self.var_bool_to_int else np.int64
                    if not isinstance(self.var_nan[col], rdt):
                        fnc_kw['nan'] = self.nan_dtype_map[rdt]
                        self.logger.warning(
                            'Chosen nan for col "{col}" not of type "{type}"; reverting to default nan: "{nan}".',
                            col=col, type=rdt, nan=self.nan_dtype_map[dt])
            else:
                if convert_inconsistent_nans:
                    fnc_kw['nan'] = self.nan_dtype_map[dt]
                else:
                    fnc_kw['nan'] = self.nan_default
            # apply dtype fix here
            df_[col] = df_[col].apply(fnc, **fnc_kw)

        if self.drop_dup_rec:
            # drop duplicate records
            df_.drop_duplicates(inplace=True)

        # cleaning up of string columns (1/2)
        if self.strip_string_columns:
            # strip string based columns
            for col, dt in self.var_dtype.items():
                if dt != str and dt != np.str_:
                    continue
                try:
                    df_[col] = df_[col].str.strip()
                except:
                    pass

        # cleaning up of string columns (2/2)
        if self.cleanup_string_columns:
            # strip string based columns
            for col, dt in self.var_dtype.items():
                if dt != str and dt != np.str_:
                    continue
                df_[col] = df_[col].apply(cleanup_string)

        # storage
        ds[self.store_key] = df_
        ds[self.store_key + '_cnts'] = pd.DataFrame(self._cnts).T

        return StatusCode.Success


def determine_preferred_dtype(dtype_cnt):
    """Determine preferred column data type."""
    # get sorted type counts for column
    type_cnts = dtype_cnt.most_common()
    if not type_cnts:
        return None

    # determine preferred type from types with the highest count
    type_order = {str: '0', np.float64: '1', np.int64: '2', np.bool_: '3'}
    return sorted((cnt[0] for cnt in type_cnts if cnt[1] == type_cnts[0][1]),
                  key=lambda t: type_order.get(t, t.__name__))[0]



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk209_read_big_data_itr

Created: 2017/02/17

Description:
    Macro to that illustrates how to loop over multiple (possibly large!)
    datasets in chunks.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import analysis, core_ops, process_manager, resources, ConfigObject, Chain
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk209_read_big_data_itr')

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk209_read_big_data_itr'
settings['version'] = 0

#########################################################################################

# when chunking through an input file, pick up only N lines in each iteration.
chunk_size = 5

#########################################################################################
# --- Set path of data
data_path = resources.fixture('dummy.csv')

#########################################################################################
# --- now set up the chains and links, based on configuration flags

# --- example 1: readdata loops over the input files, but no file chunking.

if settings.get('do_example1', True):
    ch = Chain('MyChain1')

    # --- a loop is set up in the chain MyChain.
    #     we iterate over (chunks of) the next file in the list until the iterator is done.
    #     then move on to the next chain (Overview)

    # --- readdata keeps on opening the next file in the file list.
    #     all kwargs are passed on to pandas file reader.
    read_data = analysis.ReadToDf(name='dflooper1', key='test1', sep='|', reader='csv', usecols=['x', 'y'])
    read_data.path = [data_path] * 3
    read_data.itr_over_files = True
    ch.add(read_data)

    # --- this serves as the break statement from this loop.
    #     if dataset test is empty, which can happen as the very last dataset by readdata,
    #     then skip the rest of this chain.
    skipper = core_ops.SkipChainIfEmpty()
    skipper.collection_set = ['test1']
    skipper.check_at_initialize = False
    skipper.check_at_execute = True
    ch.add(skipper)

    # --- do something useful with the test dataset here ...
    #     e.g. apply selections, or collect into histograms.

    # --- this serves as the continue statement of the loop. go back to start of the chain.
    #     repeater listens to readdata is there are any more datasets coming. if so, continue the loop.
    repeater = core_ops.RepeatChain()
    # repeat until readdata says halt.
    repeater.listen_to = 'chainRepeatRequestBy_' + read_data.name
    # repeat max of 10 times
    # repeater.maxcount = 10
    ch.add(repeater)

# --- example 2: readdata loops over the input files, with file chunking.

if settings.get('do_example2', True):
    ch = Chain('MyChain2')

    # --- a loop is set up in the chain MyChain.
    #     we iterate over (chunks of) the next file in the list until the iterator is done.
    #     then move on to the next chain (Overview)

    # --- readdata keeps on opening the next 4 lines of the open or next file in the file list.
    #     all kwargs are passed on to pandas file reader.
    read_data = analysis.ReadToDf(name='dflooper2', key='test2', sep='|', reader='csv', usecols=['x', 'y'],
                                  chunksize=chunk_size)
    read_data.path = [data_path] * 3
    ch.add(read_data)

    # --- this serves as the break statement from this loop.
    #     if dataset test is empty, which can happen as the very last dataset by readdata,
    #     then skip the rest of this chain.
    skipper = core_ops.SkipChainIfEmpty()
    skipper.collection_set = ['test2']
    skipper.check_at_initialize = False
    skipper.check_at_execute = True
    ch.add(skipper)

    # --- do something useful with the test dataset here ...
    #     e.g. apply selections, or collect into histograms.

    # query_set = seletions that are applies to incoming_records
    # after selections, only keep column in select_columns ('a', 'c')
    link = analysis.ApplySelectionToDf(read_key='test2', store_key='reduced_data', query_set=['x>1'])
    # Any other kwargs given to ApplySelectionToDf are passed on the the
    # pandas query() function.
    ch.add(link)

    # --- As an example, will merge reduced datasets back into a single, merged dataframe.
    concat = analysis.DfConcatenator()
    concat.read_keys = ['merged', 'reduced_data']
    concat.store_key = 'merged'
    concat.ignore_missing_input = True  # in first iteration input 'merged' is missing.
    ch.add(concat)

    # --- this serves as the continue statement of the loop. go back to start of the chain.
    repeater = core_ops.RepeatChain()
    # repeat until readdata says halt.
    repeater.listen_to = 'chainRepeatRequestBy_' + read_data.name
    # repeat max of 10 times
    # repeater.maxcount = 10
    ch.add(repeater)

# --- print contents of the datastore
overview = Chain('Overview')
pds = core_ops.PrintDs(name='End')
pds.keys = ['n_test1', 'n_sum_test1', 'n_test2', 'n_sum_test2', 'test2', 'n_merged']
overview.add(pds)

#########################################################################################

logger.debug('Done parsing configuration file esk209_read_big_data_itr')



"""Project: Eskapade - A python-based package for data analysis.

Class: ApplyFuncToDf

Created: 2016/11/08

Description:
    Algorithm to apply one or more functions to a (grouped) dataframe column
    or to an entire dataframe.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import collections

from eskapade import process_manager, DataStore, Link, StatusCode


class ApplyFuncToDf(Link):
    """Apply functions to data-frame.

    Applies one or more functions to a (grouped) dataframe column or an
    entire dataframe.  In the latter case, this can be done row wise or
    column wise.  The input dataframe will be overwritten.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str read_key: data-store input key
        :param str store_key: data-store output key
        :param list apply_funcs: functions to apply (list of dicts)
          - 'func': function to apply
          - 'colout' (string): output column
          - 'colin' (string, optional): input column
          - 'entire' (boolean, optional): apply to the entire dataframe?
          - 'args' (tuple, optional): args for 'func'
          - 'kwargs' (dict, optional): kwargs for 'func'
          - 'groupby' (list, optional): column names to group by
          - 'groupbyColout' (string) output column after the split-apply-combine combination
        :param dict add_columns: columns to add to output (name, column)
        """
        Link.__init__(self, kwargs.pop('name', 'apply_func_to_dataframe'))

        # process keyword arguments
        self._process_kwargs(kwargs, read_key='', store_key='', apply_funcs=[], add_columns=None)
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link."""
        self.check_arg_vals('read_key')
        if not self.apply_funcs:
            self.logger.warning('No functions to apply')

        return StatusCode.Success

    def execute(self):
        """Execute the link."""
        ds = process_manager.service(DataStore)
        assert self.read_key in ds, 'key "{key}" not in DataStore.'.format(key=self.read_key)
        df = ds[self.read_key]

        for arr in self.apply_funcs:
            # get func input
            keys = list(arr.keys())
            assert 'func' in keys, 'function input is insufficient.'
            func = arr['func']
            self.logger.debug('Applying function {function!s}.', function=func)
            args = ()
            kwargs = {}
            if 'kwargs' in keys:
                kwargs = arr['kwargs']
            if 'args' in keys:
                args = arr['args']

            # apply func
            if 'groupby' in keys:
                groupby = arr['groupby']
                if 'groupbyColout' in keys:
                    kwargs['groupbyColout'] = arr['groupbyColout']
                df = self.groupbyapply(df, groupby, func, *args, **kwargs)
            elif 'store_key' in keys:
                if 'entire' in keys:
                    result = func(df, *args, **kwargs)
                elif 'colin' in keys:
                    colin = arr['colin']
                    assert colin in df.columns
                    result = df[colin].apply(func, args=args, **kwargs)
                else:
                    result = df.apply(func, args=args, **kwargs)
                ds[arr['store_key']] = result
            else:
                assert 'colout' in keys, 'function input is insufficient'
                colout = arr['colout']
                if 'entire' in keys:
                    df[colout] = func(df, *args, **kwargs)
                elif 'colin' in keys:
                    colin = arr['colin']
                    if isinstance(colin, list):
                        for c in colin:
                            assert c in df.columns
                    else:
                        assert colin in df.columns
                    df[colout] = df[colin].apply(func, args=args, **kwargs)
                else:
                    df[colout] = df.apply(func, args=args, **kwargs)

        # add columns
        if self.add_columns is not None:
            for k, v in self.add_columns.items():
                df[k] = v

        if self.store_key is None:
            ds[self.read_key] = df
        else:
            ds[self.store_key] = df

        return StatusCode.Success

    def add_apply_func(self, func, out_column, in_column='', *args, **kwargs):
        """Add function to be applied to dataframe."""
        # check inputs
        if not isinstance(func, collections.Callable):
            self.logger.fatal('Specified function object is not callable.')
            raise AssertionError('functions in ApplyFuncToDf link must be callable objects')
        if not isinstance(out_column, str) or not isinstance(in_column, str):
            self.logger.fatal('Types of specified column names are "{in_type}" and "{out_type}."',
                              in_type=type(out_column).__name__, out_type=type(in_column).__name__)
            raise TypeError('Column names in ApplyFuncToDf must be strings.')
        if not out_column:
            self.logger.fatal('No output column specified.')
            raise RuntimeError('An output column must be specified to apply function in ApplyFuncToDf.')

        # add function
        if in_column == '':
            self.apply_funcs.append({'func': func, 'colout': out_column, 'args': args, 'kwargs': kwargs})
        else:
            self.apply_funcs.append({'colin': in_column, 'func': func, 'colout': out_column, 'args': args,
                                     'kwargs': kwargs})

    def groupbyapply(self, df, groupby_columns, applyfunc, *args, **kwargs):
        """Apply groupby to dataframe."""
        if 'groupbyColout' not in kwargs:
            return df.groupby(groupby_columns).apply(applyfunc, *args, **kwargs).reset_index(drop=True)
        else:
            colout = kwargs['groupbyColout']
            kwargs.pop('groupbyColout')
            t = df.groupby(groupby_columns).apply(applyfunc, *args, **kwargs)
            for _ in range(0, len(groupby_columns)):
                t.index = t.index.droplevel()
            df[colout] = t
            return df



"""Project: Eskapade - A python-based package for data analysis.

Class: DsApply

Created: 2018-06-30

Description:
    Simple link to execute functions, to which datastore has been passed.

    Helps in the development of links.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import process_manager, ConfigObject, DataStore, Link, StatusCode


class DsApply(Link):

    """Simple link to execute functions to which datastore is passed."""

    def __init__(self, **kwargs):
        """Initialize an instance.

        :param str name: name of link
        :param list apply: list of functions to execute at execute(), to which datastore is passed
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'DsApply'))

        # Process and register keyword arguments. If the arguments are not given, all arguments are popped from
        # kwargs and added as attributes of the link. Otherwise, only the provided arguments are processed.
        self._process_kwargs(kwargs, apply=[])

        # check residual kwargs; exit if any present
        self.check_extra_kwargs(kwargs)

    def execute(self):
        """Execute the link.

        :returns: status code of execution
        :rtype: StatusCode
        """
        # --- your algorithm code goes here
        self.logger.debug('Now executing link: {link}.', link=self.name)

        ds = process_manager.service(DataStore)

        # --- pass ds to list of functions, to execute bits of code
        #     by doing this here, the objects previously created can be picked up.
        for func in self.apply:
            self.logger.debug('Now executing function: {function}.', function=func.__name__)
            func(ds)

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Class: Break

Created: 2017/02/26

Description:
    Algorithm to send break signal to process manager and halt execution

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import Link
from eskapade import StatusCode


class Break(Link):
    """Halt execution.

    Link sends failure signal and halts execution of process manager.  Break
    the execution of the processManager at a specific location by simply
    adding this link at any location in a chain.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'Break'))

        # check residual kwargs. exit if any present
        self.check_extra_kwargs(kwargs)

    def execute(self):
        """Execute the link."""
        # halt the execution of process_manager by sending a failure signal
        self.logger.info('Now sending break signal to halt execution!')
        return StatusCode.Failure



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/02/27

Description:
    Base class and core implementations of run-process services

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import os
import pickle
import re
from collections import defaultdict
from typing import Any

import eskapade.utils
from eskapade.core.definitions import CONFIG_DEFAULTS
from eskapade.core.definitions import CONFIG_OPTS_SETTERS
from eskapade.core.definitions import CONFIG_VARS
from eskapade.core.definitions import USER_OPTS
from eskapade.core.exceptions import UnknownSetting
from eskapade.logger import Logger


class ProcessServiceMeta(type):
    """Meta class for process-services base class."""

    def __str__(self):
        """Get printable specification of service."""
        return '{0:s}.{1:s}'.format(self.__module__, self.__name__)

    @property
    def persist(self):
        """Flag to indicate if service can be persisted."""
        return self._persist


class ProcessService(metaclass=ProcessServiceMeta):
    """Base class for process services."""

    logger = Logger()
    _persist = False

    def __init__(self):
        """Initialize service instance."""
        pass

    def __str__(self):
        """Get printable specification of service instance."""
        return '{0!s} ({1:s})'.format(type(self), hex(id(self)))

    @classmethod
    def create(cls):
        """Create an instance of this service.

        :returns: service instance
        :rtype: ProcessService
        """
        # create instance and make sure the service is initialized
        inst = cls()
        ProcessService.__init__(inst)
        return inst

    def finish(self):
        """Finish current processes.

        This function can be implemented by a process-service implementation to
        finish running processes and clean up to prepare for a reset of the
        process manager.  This would typically involve deleting large objects
        and closing files and database connections.
        """
        pass

    @classmethod
    def import_from_file(cls, file_path):
        """Import service instance from a Pickle file.

        :param str file_path: path of Pickle file
        :returns: imported service instance
        :rtype: ProcessService
        :raises: RuntimeError, TypeError
        """
        # check if service can be persisted
        if cls.persist:
            cls.logger.debug('Importing service instance of "{cls!s}" from file "{path}".', cls=cls, path=file_path)
        else:
            cls.logger.debug('Not importing service "{cls!s}".', cls=cls)
            return None

        # check specified file path
        if not os.path.isfile(file_path):
            cls.logger.fatal('Specified path for importing "{cls!s}" instance is not a file "{path}".',
                             cls=cls, path=file_path)
            raise RuntimeError('Invalid file path specified for importing process service.')

        try:
            # try to open file and import instance
            with open(file_path, 'rb') as inst_file:
                inst = pickle.load(inst_file)
        except Exception as exc:
            # re-raise exeption if import failed
            cls.logger.warning('Failed to import service instance of "{cls!s}" from file "{path}".',
                               cls=cls, path=file_path)
            raise exc

        # check type of instance
        if not isinstance(inst, cls):
            cls.logger.fatal('Expected to import "{cls!s}" instance, got object of type "{type}".',
                             cls=cls, type=type(inst).__name__)
            raise TypeError('Incorrect type for imported service object.')

        return inst

    def persist_in_file(self, file_path):
        """Persist service instance in Pickle file.

        :param str file_path: path of Pickle file
        """
        # check if service can be persisted
        if type(self).persist:
            self.logger.debug('Persisting service instance "{instance!s}" in file "{path}".',
                              instance=self, path=file_path)
        else:
            self.logger.debug('Not persisting service "{type!s}".', type=type(self))
            return

        try:
            # try to persist
            with open(file_path, 'wb') as inst_file:
                pickle.dump(self, inst_file)
        except Exception as exc:
            # give warning if persisting failed
            self.logger.warning('Failed to persist service instance "{instance!s}" in file "{path}".',
                                instance=self, path=file_path)
            self.logger.warning('Caught exception "{exc!s}".', exc=exc)


class ConfigObject(ProcessService):
    """Configuration settings for Eskapade.

    The ConfigObject is a dictionary meant for containing global settings of
    Eskapade.  Settings are set in the configuration macro of an
    analysis, or on the command line.

    The ConfigObject is a dictionary meant only for storing global settings
    of Eskapade.  In general, it is accessed through the process
    manager.

    Example usage:

    >>> # first set logging output level.
    >>> from eskapade.logger import Logger, LogLevel
    >>> logger = Logger()
    >>> logger.log_level = LogLevel.DEBUG

    Obtain the ConfigObject from any location as follows:

    >>> from eskapade import process_manager
    >>> from eskapade import ConfigObject
    >>> settings = process_manager.service(ConfigObject)

    One can treat the ConfigObject as any other dictionary:

    >>> settings['foo'] = 'bar'
    >>> foo = settings['foo']

    Write the ConfigObject to a pickle file with:

    >>> settings.persist_in_file(file_path)

    And reload from the pickle file with:

    >>> settings = ConfigObject.import_from_file(file_path)

    A ConfigObject pickle file can be read in by Eskapade with the
    command line option (-u).
    """

    _persist = True

    def __init__(self):
        """Initialize ConfigObject instance."""
        self.__settings = defaultdict()

        # Initialize self with default values.
        for section in CONFIG_VARS.values():
            for config in section:
                self.__settings[config] = CONFIG_DEFAULTS.get(config)

        # initialize batch-mode setting with display variable from environment
        display = eskapade.utils.get_env_var('display')
        self.__settings['batchMode'] = display is None or not re.search(':\d', display)

    def __repr__(self):
        return repr(self.__settings)

    def __getitem__(self, setting: str) -> Any:
        """Get value of setting by name.

        :param setting: The setting to get.
        :return: The value of setting.
        :raise: UnknownSetting if it does not exist.
        """
        if setting in self.__settings:
            return self.__settings[setting]

        raise UnknownSetting('Unknown setting {setting}!'.format(setting=setting))

    def __setitem__(self, setting: str, value: Any) -> None:
        """Set the value of a setting.

        Note this overrides the current value a the setting.

        :param setting:
        :param value:
        :return: None
        """
        self.__settings[setting] = value

    def __contains__(self, setting):
        """Check if a setting is present.

        :param setting: The setting to check for.
        :return: True if setting is present else False.
        """
        return setting in self.__settings

    def get(self, setting: str, default: Any = None) -> object:
        """Get value of setting. If it does not exists return the default value.

        :param setting: The setting to get.
        :param default: The default value of the setting.
        :return: The value of the setting or None if it does not exist.
        """
        try:
            return self.__getitem__(setting)
        except UnknownSetting:
            return default

    def __copy__(self):
        """Perform a shallow copy of self.

        :return: copy
        """
        clone = ConfigObject()
        clone.__settings = self.__settings.copy()

        return clone

    def copy(self):
        """Perform a shallow copy of self.

        :return: copy
        """
        return self.__copy__()

    def io_base_dirs(self) -> dict:
        """Get configured base directories.

        :returns: base directories
        :rtype: dict
        """
        return dict([(key + '_dir', self.__getitem__(key + 'Dir'))
                     for key in ['config', 'results', 'data', 'macros', 'templates']])

    def io_conf(self):
        """Get I/O configuration.

        The I/O configuration contains storage locations and basic analysis
        info.

        :return: I/O configuration
        :rtype: IoConfig
        """
        return ConfigObject.IoConfig(analysis_name=self['analysisName'],
                                     analysis_version=self['version'],
                                     **self.io_base_dirs())

    def Print(self):
        """Print a summary of the settings."""
        # print standard settings
        self.logger.info('Run configuration')
        for sec, sec_keys in CONFIG_VARS.items():
            if not sec_keys:
                continue
            self.logger.info('  {config}:', config=sec)
            max_key_len = max(len(k) for k in sec_keys)
            for key in sec_keys:
                self.logger.info('    {{key:<{:d}s}}  {{value!s}}'.format(max_key_len), key=key, value=self.get(key))

        # print additional custom settings
        add_keys = sorted(set(self.__settings.keys()) - set(o for s in CONFIG_VARS.values() for o in s))
        if add_keys:
            self.logger.info('  custom:')
            max_key_len = max(len(k) for k in add_keys)
            for key in add_keys:
                self.logger.info('    {{0:<{:d}s}}  {{1!s}}'.format(max_key_len).format(key, self.get(key)))

    def add_macros(self, macro_paths):
        """Add configuration macros for Eskapade run."""
        # convert input to list if single path is specified
        if isinstance(macro_paths, str):
            macro_paths = [macro_paths]

        # loop over specified file paths
        # FIXME: add all specified macros instead of only the first one
        for path in macro_paths:
            self['macro'] = path
            break

    def set_user_opts(self, parsed_args):
        """Set options specified by user on command line.

        :param argparse.Namespace parsed_args: parsed user arguments
        """
        # loop over arguments
        args = vars(parsed_args)
        known_opts = set(opt for sec_opts in USER_OPTS.values() for opt in sec_opts)
        # only process known config options
        for opt_key in set(args).intersection(known_opts):
            # call setter function for this user option
            CONFIG_OPTS_SETTERS[opt_key](opt_key, self, args)

    class IoConfig(dict):
        """Configuration object for I/O operations."""

        _conf_items = dict(analysis_name=str,
                           analysis_version=None,
                           results_dir=str,
                           data_dir=str,
                           macros_dir=str,
                           templates_dir=str)

        def __init__(self, **input_config):
            """Initialize IoConfig instance."""
            # check required items
            for key, val_type in self._conf_items.items():
                if key not in input_config:
                    ConfigObject.logger.fatal('Item "{key}" not found in input IO configuration.', key=key)
                    raise KeyError('Missing item(s) in input configuration for IoConfig.')
                if val_type and not isinstance(input_config[key], val_type):
                    ConfigObject.logger.fatal('Item "{key}" has type "{type}" ("{name}" required).',
                                              key=key, type=type(input_config[key]).__name__, name=str.__name__)
                    raise TypeError('Incorrect type for item(s) in input configuration for IoConfig.')

            # initialize dictionary
            dict.__init__(self, **input_config)
            self['analysis_version'] = str(self['analysis_version'])


class DataStore(ProcessService, dict):
    """Store for transient data sets and related objects.

    The data store is a dictionary meant for storing transient data sets or
    any other objects.  Links can take one or several data sets as input,
    transform them or use them as input for a model, and store the output
    back again in the datastore, to be picked up again by any following link.

    Example usage:

    >>> # first set logging output level.
    >>> from eskapade.logger import Logger, LogLevel
    >>> logger = Logger()
    >>> logger.log_level = LogLevel.DEBUG

    Obtain the global datastore from any location as follows:

    >>> from eskapade import process_manager
    >>> from eskapade import DataStore
    >>>
    >>>
    >>> ds = process_manager.service(DataStore)

    One can treat the datastore as any other dict:

    >>> ds['a'] = 1
    >>> ds['b'] = 2
    >>> ds['0'] = 3
    >>> a = ds['a']

    Write the datastore to a pickle file with:

    >>> ds.persist_in_file(file_path)

    And reload from the pickle file with:

    >>> ds = DataStore.import_from_file(file_path)
    """

    _persist = True

    def Print(self):
        """Print a summary the data store contents."""
        self.logger.info('Summary of data store ({n:d} objects)', n=len(self))
        if not self:
            return

        max_key_len = max(len(k) for k in self.keys())
        for key in sorted(self.keys()):
            self.logger.info('  {{0:<{:d}s}}  <{{1:s}}.{{2:s}} at {{3:x}}>'.
                             format(max_key_len).format(key,
                                                        type(self[key]).__module__,
                                                        type(self[key]).__name__,
                                                        id(self[key])))



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/09/14

Description:

    A collection of (generic) meta classes for some (design) patterns:

        - Singleton: Meta class for the Singleton pattern.
        - Processor: Meta class with abstract methods initialize,
          execute, and finalize.
        - ProcessorSequence: A simple (processor) sequence container.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from abc import ABCMeta, abstractmethod
from weakref import proxy

from eskapade.core.definitions import StatusCode
from eskapade.logger import Logger


class Singleton(type):
    """Metaclass for singletons.

    Any instantiation of a Singleton class yields the exact same object, e.g.:

    >>> class Klass(metaclass=Singleton):
    >>>     pass
    >>>
    >>> a = Klass()
    >>> b = Klass()
    >>> a is b
    True

    See https://michaelgoerz.net/notes/singleton-objects-in-python.html.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Return the singleton."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]

    @classmethod
    def __instancecheck__(mcs, instance) -> bool:
        if instance.__class__ is mcs:
            return True
        return isinstance(instance.__class__, mcs)


class Processor(metaclass=ABCMeta):
    """Processor metaclass."""

    logger = Logger()  # type: Logger

    def __init__(self, name: str):
        """Initialize the Processor object."""
        super().__init__()
        name = name or self.__class__.__name__
        self.__name = name  # type: str
        self.__hash = None  # type: int
        self.__parent = None

    def __str__(self) -> str:
        return self.__name

    def __repr__(self) -> str:
        return '<{klass!s} name={name!s} parent={parent!r} id={id!s}>'.format(klass=self.__class__.__name__,
                                                                              name=self.name,
                                                                              parent=self.__parent,
                                                                              id=id(self))

    def __eq__(self, other: 'Processor') -> bool:
        return isinstance(other, type(self)) and self.__name == other.__name

    def __hash__(self) -> int:
        if self.__hash is None:
            self.__hash = hash((type(self), self.__name))

        return self.__hash

    def _initialize(self):
        """Wrapper to call user implemented initialize."""
        self.logger.debug('Initializing link "{link!s}".', link=self)

        status = self.initialize()

        if status == StatusCode.Success:
            self.logger.debug('Successfully initialized link "{link!s}".', link=self)

        return status

    @abstractmethod
    def initialize(self):
        """Initialization logic for processor."""
        raise NotImplementedError

    def _execute(self):
        """Wrapper to call user implemented execute."""
        self.logger.debug('Executing link "{link!s}".', link=self)

        status = self.execute()

        if status == StatusCode.Success:
            self.logger.debug('Successfully executed link "{link!s}".', link=self)

        return status

    @abstractmethod
    def execute(self):
        """Execution logic for processor."""
        raise NotImplementedError

    def _finalize(self):
        """Wrapper to call user implemented finalize."""
        self.logger.debug('Finalizing link "{link!s}".', link=self)

        status = self.finalize()

        if status == StatusCode.Success:
            self.logger.debug('Successfully finalized link "{link!s}".', link=self)

        return status

    @abstractmethod
    def finalize(self):
        """Finalization logic for processor."""
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Get the name of processor.

        :return: The name of the processor.
        :rtype: str
        """
        return self.__name

    @property
    def parent(self):
        """Get the group parent.

        :return: The parent/group processor sequence.
        """
        return self.__parent

    @parent.setter
    def parent(self, the_parent) -> None:
        """Set the group parent.

        :param the_parent: The parent/group processor sequence.
        """
        self.__parent = None

        if the_parent is not None:
            # The parent will most likely outlive the processor
            # and therefore we do not want keep a strong reference
            # to the parent.
            self.__parent = proxy(the_parent)


class _ProcessorNode(object):
    """A Processor Node.

    This is used by :class:`ProcessingSequence` to doubly link processors that are
    to be executed sequentially.

    :attr prev: The previous processor in the chain.
    :attr next: The next processor in the chain.
    :attr value: The current processor.
    """

    __slots__ = 'prev', 'next', 'value', '__weakref__'

    def __init__(self):
        """Initialize the ProcessorNode object."""
        self.prev = None  # type: Processor
        self.next = None  # type: Processor
        self.value = None  # type: Processor


class ProcessorSequence(object):
    """A doubly linked processor sequence.

    It remembers the order in which processors are added to the sequence.
    It also checks if a processor already has been added to the sequence.
    """

    __slots__ = '__end', '__map', '__weakref__'

    def __init__(self):
        """Initialize the ProcessorSequence object."""
        super().__init__()
        self.__end = end = _ProcessorNode()  # type: _ProcessorNode
        end.prev = end.next = end  # type: _ProcessorNode
        self.__map = {}

    def __len__(self) -> int:
        """Get the length, i.e. the number of processors.

        :return: The length of the sequence.
        :rtype: int
        """
        return len(self.__map)

    def __contains__(self, processor: Processor) -> bool:
        """Check if sequence contains processor.

        :param processor: The processor to check.
        :type processor: Processor
        :return: True when the sequence contains the processor otherwise False.
        :rtype: bool
        """
        return processor in self.__map

    def __iter__(self):
        """Forward iterator.

        :return: Generator object.
        """
        end = self.__end
        curr_node = end.next
        while curr_node is not end:
            yield curr_node.value
            curr_node = curr_node.next

    def __reversed__(self):
        """Reverse iterator.

        :return: Generator object.
        """
        end = self.__end
        curr_node = end.prev
        while curr_node is not end:
            yield curr_node.value
            curr_node = curr_node.prev

    def __repr__(self) -> str:
        return '<{klass!s} id={id!s} sequence={repr!s}>'.format(klass=self.__class__.__name__,
                                                                id=id(self),
                                                                repr=set(self) if self else '{}')

    def add(self, processor: Processor) -> None:
        """Add a processor to the sequence.

        :param processor: The processor to add.
        :type processor: Processor
        :raise KeyError: When a processor of the same type and name already exists.
        """
        if processor not in self.__map:
            self.__map[processor] = node = _ProcessorNode()
            end = self.__end
            last = end.prev
            node.prev, node.next, node.value = last, end, processor
            last.next = end.prev = proxy(node)
        else:
            raise KeyError('Processor "{processor!r}" already exists!'.format(processor=processor))

    def discard(self, processor: Processor) -> None:
        """Remove a processor from the sequence.

        :param processor: The processor to remove.
        :type processor: Processor
        :raise KeyError: When the processor does not exist.
        """
        if processor in self.__map:
            node = self.__map.pop(processor)
            node.prev.next = node.next
            node.next.prev = node.prev
        else:
            raise KeyError('Unknown processor "{processor!r} to discard"!'.format(processor=processor))

    def pop(self, last: bool = True) -> Processor:
        """Return the popped processor. Raise KeyError if empty.

        By default a processor is popped from the end of the sequence.

        :param last: Pop processor from the end of the sequence. Default is True.
        :type last: bool
        :return: The pop processor.
        :raise KeyError: When trying to pop from an empty list.
        """
        if not self:
            raise KeyError('Set is empty!')
        # noinspection PyTypeChecker
        processor = next(reversed(self)) if last else next(iter(self))
        self.discard(processor)
        return processor

    def clear(self) -> None:
        """Clear the sequence."""
        # Reset end node.
        self.__end.prev = self.__end.next = self.__end
        # Clear processor to processor node map.
        self.__map.clear()



"""Project: Eskapade - A python-based package for data analysis.

Created: 2016/11/08

Description:
    Eskapade exceptions.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""


class Error(Exception):
    """Base class for all Eskapade core exceptions."""


class UnknownSetting(Error):
    """The user requested an unknown setting."""



"""Project: Eskapade - A python-based package for data analysis.

Class: ApplySelectionToDf

Created: 2016/11/08

Description:
    Algorithm to apply queries to input dataframe

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy

import pandas as pd

from eskapade import process_manager, DataStore, Link, StatusCode


class ApplySelectionToDf(Link):
    """Applies queries with sub-selections to a pandas dataframe."""

    def __init__(self, **kwargs):
        """Initialize link instance.

        Input dataframe is not overwritten, unless instructed to do so in kwargs.

        :param str name: name of link
        :param str read_key: key of data to read from data store
        :param str store_key: key of data to store in data store. If not set read_key is overwritten.
        :param list query_set: list of strings, query expressions to evaluate in the same order,
            see pandas documentation
        :param list select_columns: column names to select after querying
        :param bool continue_if_failure: if True continues with next query after failure (optional)
        :param kwargs: all other key word arguments are passed on to the pandas queries.
        """
        Link.__init__(self, kwargs.pop('name', 'ApplySelectionToDf'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             read_key='',
                             store_key=None,
                             query_set=[],
                             select_columns=[],
                             continue_if_failure=False)

        # pass on remaining kwargs to pandas query
        self.kwargs = copy.deepcopy(kwargs)

    def initialize(self):
        """Initialize the link.

        Perform checks on provided attributes.
        """
        # query_set or select_columns needs to be set.
        assert len(self.read_key) > 0, 'read_key has not been set.'

        if self.store_key is None:
            self.logger.warning('store_key has not been set, now set to: {key}', key=self.read_key)
            self.store_key = self.read_key
        else:
            assert isinstance(self.store_key, str) and len(self.store_key), 'store_key has not been set.'

        if isinstance(self.query_set, str):
            self.query_set = [self.query_set]
        elif not isinstance(self.query_set, list):
            raise Exception('Query set is not a list of strings.')

        if isinstance(self.select_columns, str):
            self.select_columns = [self.select_columns]
        elif not isinstance(self.select_columns, list):
            raise Exception('Column selection is not a list of strings.')

        assert len(self.query_set) or len(self.select_columns), 'No selections have been provided.'

        self.logger.info('kwargs passed on to pandas query function are: {kwargs}', kwargs=self.kwargs)

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Applies queries or column selection to a pandas DataFrame.
        Input dataframe is not overwritten, unless told to do so in kwargs.

        1. Apply queries, in order of provided query list.
        2. Select columns (if provided).
        """
        ds = process_manager.service(DataStore)
        assert self.read_key in ds, 'Key "{key}" not in DataStore.'.format(key=self.read_key)
        assert isinstance(ds[self.read_key],
                          pd.DataFrame), 'Object with key {} is not a pandas DataFrame.'.format(self.read_key)

        # 1. apply queries to input dataframe.
        #    input dataframe is not overwritten, unless told to do so in kwargs.
        do_continue = True
        if len(self.query_set):
            # apply first query
            query = self.query_set[0]
            try:
                df = ds[self.read_key].query(query, **self.kwargs)
            except Exception:
                if not self.continue_if_failure:
                    raise ValueError('Failed to apply query <{}> to dataframe <{}>.'.format(query, self.read_key))
                else:
                    orig_df_cols = (ds[self.read_key]).columns
                    df = pd.DataFrame(columns=orig_df_cols)
                    do_continue = False
            # apply rest of the queries if any
            if do_continue:
                for query in self.query_set[1:]:
                    try:
                        df = df.query(query, **self.kwargs)
                    except Exception:
                        if not self.continue_if_failure:
                            raise ValueError(
                                'Failed to apply query <{}> to dataframe <{}>.'.format(query, self.read_key))
                        else:
                            orig_df_cols = (ds[self.read_key]).columns
                            df = pd.DataFrame(columns=orig_df_cols)
                            break

        # 2. apply column selection to input dataframe.
        #    input dataframe is not overwritten.
        if len(self.select_columns):
            if 'df' not in vars():
                df = (ds[self.read_key]).copy(deep=False)
            try:
                df = df[self.select_columns]
            except Exception:
                if not self.continue_if_failure:
                    raise ValueError(
                        'Failed to select columns <{self.select_columns!s}> of dataframe <{self.read_key}>.'
                        .format(self=self))
                else:
                    df = pd.DataFrame(columns=self.select_columns)

        assert 'df' in vars(), 'No dataframe available for storage?'

        ds[self.store_key] = df
        ds['n_' + self.store_key] = len(df.index)

        self.logger.info('Stored dataframe with key <{key}> and length <{length:d}>.',
                         key=self.store_key, length=len(df.index))

        return StatusCode.Success
